"""RunnableManager classs"""

import a99
import threading
from PyQt5.QtCore import QObject, pyqtSignal
import multiprocessing
import copy
import time
from threading import Lock
import collections
import pyfant

__all__ = ["RunnableManager"]



def _tm_print(s):
    """print for the thread manager2 (debugging)."""
    print(("^^ %s ^^" % s))


class RunnableManagerError(Exception):
    pass


@a99.froze_it
class RunnableManager(QObject, threading.Thread):
    """
    Thread takes care of other threads.

    Keyword arguments:
      max_simultaneous=multiprocessing.cpu_count()
      flag_auto_clean=False: if set, will load result and remove the session
       directory as soon as a runnable has finished
    """

    # Emitted when a new thread is added
    runnable_added = pyqtSignal()

    # Emitted whenever the state of any thread has changed
    runnable_changed = pyqtSignal()

    # Emitted when there are no runnables left to run
    finished = pyqtSignal()

    @property
    def num_finished(self):
        with self.__lock:
            return self.__num_finished

    @property
    def time_finished(self):
        with self.__lock:
            return self.__time_finished

    @property
    def num_runnables(self):
        with self.__lock:
            return len(self.__runnables)

    @property
    def num_failed(self):
        with self.__lock:
            return self.__num_failed

    @property
    def time_started(self):
        with self.__lock:
            return self.__time_started

    @property
    def max_simultaneous(self):
        with self.__lock:
            return self.__max_simultaneous

    @property
    def flag_cancelled(self):
        with self.__lock:
            return self.__flag_cancelled

    @property
    def flag_exited(self):
        with self.__lock:
            return self.__time_started is not None and not self.is_alive()

    @property
    def flag_finished(self):
        with self.__lock:
            return self.__num_finished == len(self.__runnables)

    @property
    def flag_start_called(self):
        with self.__lock:
            return self.__time_started is not None

    @property
    def flag_paused(self):
        with self.__lock:
            return self.__flag_paused

    @property
    def flag_failed(self):
        with self.__lock:
            return self.__num_failed > 0

    @property
    def flag_auto_clean(self):
        with self.__lock:
            return self.__flag_auto_clean

    @property
    def flag_success(self):
        """Success is defined as nothing left to run and no fails."""
        with self.__lock:
            return self.__num_failed == 0 and len(self.__idxs_to_run) == 0

    @property
    def time_per_runnable(self):
        with self.__lock:
            return self.__time_per_runnable

    @property
    def max_simultaneous(self):
        """Maximum runnables running simultaneously. Default is multiprocessing.cpu_count()"""
        return self.__max_simultaneous

    @max_simultaneous.setter
    def max_simultaneous(self, x):
        self.__max_simultaneous = x

    @property
    def flag_verbose(self):
        """Whether to log progress-related messages"""
        return self.__max_simultaneous

    @flag_verbose.setter
    def flag_verbose(self, x):
        self.__flag_verbose = x


    def __init__(self, *args, max_simultaneous=None, flag_auto_clean=False, flag_verbose=False,
                 flag_exit_if_fail=False, **kwargs):
        self.__max_simultaneous = max_simultaneous
        self.__flag_auto_clean = flag_auto_clean
        self.__flag_verbose = flag_verbose
        self.__flag_exit_if_fail = flag_exit_if_fail
        if self.__max_simultaneous is None: self.__max_simultaneous = multiprocessing.cpu_count()
        QObject.__init__(self)
        threading.Thread.__init__(self, *args, **kwargs)
        self.__logger = a99.get_python_logger()
        # counts finished
        self.__num_finished = 0
        # counts failed
        self.__num_failed = 0
        # all runnables
        self.__runnables = []
        # FIFO stack containing indexes of __runnables to run
        self.__idxs_to_run = collections.deque()
        # flag to exit as soon as possible
        self.__flag_exit = False
        # set to True if explicitly cancelled through calling cancel()
        self.__flag_cancelled = False
        # indicates whether the runnable manages is paused.
        # When paused, it will not delegate new tasks to the runners.
        self.__flag_paused = False

        # Runner threads
        self.__runners = []

        # # Locks
        self.__lock = Lock()
        # self.__lock = MyLock("RM Lock", True)

        # # Statistics
        # time the thread has started
        self.__time_started = None
        # time the last runnable has finished
        self.__time_finished = None
        # average time to run each runnable
        self.__time_per_runnable = 0
        
        for i in range(self.__max_simultaneous):
            t = _Runner(self)
            self.__runners.append(t)

    def start(self):
        self.__time_started = time.time()
        threading.Thread.start(self)

    def cancel(self):
        with self.__lock:
            self.__flag_cancelled = True
            self.__flag_exit = True

    def pause(self):
        """Pauses the delegation of runnables to runners.

        *Attention*: messes with time estimate."""
        with self.__lock:
            self.__flag_paused = True

    def resume(self):
        with self.__lock:
            self.__flag_paused = False

    def get_runnables_copy(self):
        """Returns a copy of self.__runnables.

        The result is a copy of the self.__runnables list, but the elements are
        the actual thread objects."""
        with self.__lock:
            return copy.copy(self.__runnables)

    def get_times(self):
        """Returns (ellapsed, total, remaining)."""
        with self.__lock:
            return self.__unlocked_get_times()

    def add_runnables(self, runnables):
        with self.__lock:
            self.__unlocked_add_runnables(runnables)
        self.runnable_added.emit()

    def exit(self):
        self.__flag_exit = True
        
    def kill_runnables(self):
        with self.__lock:
            for runner in self.__runners:
                if runner.runnable and runner.runnable.flag_running:
                    runner.kill_runnable()

    def kill_runnable(self, runnable):
        with self.__lock:
            self.__unlocked_kill_runnable(runnable)

    def get_summary_report(self):
        """Returns list with information such as ellapsed time, remaining time etc."""

        with self.__lock:
            l = []
            l.append("***finished: %d/%d" % (self.__num_finished, len(self.__runnables)))
            if self.__num_failed > 0:
                l.append("***failed: %d" % (self.__num_failed))
            if self.__time_started:
                ella, tot, rema = self.__unlocked_get_times()
                l.append("***time ellapsed: %s" % a99.seconds2str(ella))
                if self.__num_finished > 0:
                    tpr = self.__time_per_runnable
                    l.append("***time per runnable: %s" % a99.seconds2str(tpr))
                    l.append("***time total estimate: %s" % a99.seconds2str(tot))
                    l.append("***time remaining estimate: %s" % a99.seconds2str(rema))
        return l

    def __str__(self):
        l, temp = [], []  # string list, temporary list
        with self.__lock:
            # loop to determine name width
            w = 0
            for i, t in enumerate(self.__runnables):
                s_title = t.conf.sid.dir
                if s_title is None:
                    s_title = '...'
                w = max(w, len(s_title))
                status = t.get_status()
                s_status = "?" if status is None else str(status)
                #else:
                #    s_prog = 'finished' if t.flag_finished else '-'
                temp.append((i, s_title, s_status))

            for i, name, progress in temp:
                l.append(("%03d %-"+str(w)+"s %s") % (i, name, progress))
        l.extend(self.get_summary_report())

        return "\n".join(l)

    def wait_until_finished(self):
        if not self.is_alive():
            raise RuntimeError("Runnable manager not running")
        while True:
            if self.flag_finished:
                self.exit()
                break
            if self.__flag_verbose:
                self.__logger.info("\n".join(self.get_summary_report()))
            time.sleep(1)

    def retry_failed(self):
        """Retries all failed runnables."""
        if not self.flag_finished:
            raise RuntimeError("Can only retry when finished!")
        with self.__lock:
            temp = self.__num_failed
            self.__num_failed = 0
            self.__num_finished -= temp
            to_add = [x for x in self.__runnables if not x.flag_success]
            for runnable in to_add:
                runnable.reset()
                self.__runnables.remove(runnable)
            self.__unlocked_add_runnables(to_add)
        self.runnable_added.emit()

    def _finish(self, runner):
        """Called by a _Runner to inform that a runnable has finished.

        This is for monitoring purpose only and has no effect in the workings
        of the thread manager.
        """

        flag_exit = False

        with self.__lock:
            if not runner.runnable.flag_success:
                self.__num_failed += 1
                if self.__flag_exit_if_fail:
                    flag_exit = True

            self.__num_finished += 1
            t = time.time()
            self.__time_per_runnable = (t-self.__time_started)/self.__num_finished
            if self.__num_finished == len(self.__runnables):
                self.__time_finished = t

        self.runnable_changed.emit()

        if self.flag_finished:
            self.finished.emit()

        if flag_exit:
            self.exit()

    def run(self):
        flag_sleep = False
        it = 0  # thread index
        self.__logger.debug("Will run %d runnables" % len(self.__runnables))
        while True:
            if self.__flag_exit:
                for runner in self.__runners:
                    if runner.runnable and runner.runnable.flag_running:
                        runner.kill_runnable()
                    if runner.is_alive():
                        runner.exit()
                break

            if self.__flag_paused:
                flag_sleep = True
            else:
                with self.__lock:
                    if len(self.__idxs_to_run) == 0:
                        # No new tasks for the moment
                        flag_sleep = True
                    else:
                        idx_to_run = self.__idxs_to_run[0]
                        runnable = self.__runnables[idx_to_run]
                        j = 0
                        # loop to find idle thread
                        while True:
                            if j >= self.__max_simultaneous:
                                # gave a full turn without finding idle thread
                                flag_sleep = True
                                break
                            r = self.__runners[it]

                            it += 1
                            if it >= self.__max_simultaneous:
                                it = 0

                            if r.flag_idle:
                                r.set_runnable(runnable)
                                self.__logger.debug("Assigned #%d :%s to %s" %
                                 (idx_to_run, runnable.name, r.name))
                                self.__idxs_to_run.popleft()

                                if not r.is_alive():
                                    r.start()
                                break
                            j += 1

            if flag_sleep:
                T = 0.1
                time.sleep(T)  # Chilling out for a while
                flag_sleep = False
    
        self.__logger.debug("TM exited")

    def __unlocked_add_runnables(self, runnables):
        n = len(self.__runnables)

        self.__runnables.extend(runnables)
        self.__idxs_to_run.extend(list(range(n, n + len(runnables))))

    def __unlocked_kill_runnable(self, runnable):
        flag_found = False
        with self.__lock:
            for runner in self.__runners:
                if runner.runnable == runnable and runner.runnable.flag_running:
                    flag_found = True
                    runner.kill_runnable()
                    break
        if not flag_found:
            raise RunnableManagerError("Runnable '%s' not running!" % runnable)

    def __unlocked_get_times(self):
        inf = float("inf")
        ella, tot, rema = 0, 0, 0
        nr = len(self.__runnables)
        if self.__time_started is None:
            if nr > 0:
                tot, rema = inf, inf
        else:
            ella = time.time() - self.__time_started
            if nr > 0:
                nf = self.__num_finished
                if nf == 0:
                    tot, rema = inf, inf
                else:
                    tpr = self.__time_per_runnable
                    tf = tpr * nf  # estimated time until last recorded finished
                    tot = tpr * nr
                    if not self.__flag_exit:
                        rema = tpr * (nr - nf) - (ella - tf)
                        if rema < 0:
                            rema = 0
                    else:
                        rema = inf
        return ella, tot, rema


@a99.froze_it
class _Runner(threading.Thread):
    """
    Thread that runs object of class Runnable.
    """

    def __init__(self, manager, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        assert isinstance(manager, RunnableManager)
        self.runnable = None
        self.manager = manager
        self.name = self.__class__.__name__+a99.random_name()
        self.flag_exit = False
        self.flag_idle = True
        self.lock = Lock()
        self.__logger = None


    def set_runnable(self, x):
        assert self.flag_idle or x is None
        with self.lock:
            self.runnable = x
            self.flag_idle = x is None

    def exit(self):
        """Intention to exit asap."""
        self.flag_exit = True
        self.__logger.debug("ok to exit")

    def kill_runnable(self):
        """Attempt to kill whatever is running."""
        if self.runnable:
            self.runnable.kill()

    def run(self):
        self.__logger = a99.get_python_logger()
        # this was leaving file open after finished add_file_handler(self.__logger, "python.log")
        self.__logger.debug("\\o/ %s is alive \\o/" % (self.name))
        misses = 0
        flag_sleep = False
        while True:
            if self.flag_exit:
                break
            with self.lock:
                if self.runnable:
                    self.flag_idle = False
                    try:
                        self.runnable.run()
                        if self.manager.flag_auto_clean and self.runnable.flag_success:
                            self.runnable.load_result()
                            self.runnable.sid.clean()

                    except pyfant.FailedError as E:
                        # If runnable fails, the current behaviour is use ErrorCollector and
                        # critical-log the error if found something; otherwise

                        dir_ = self.runnable.conf.sid.dir
                        k = a99.ErrorCollector(flag_warnings=False)
                        k.collect_errors(dir_)
                        template = "Caught in rm.py::_Runner.run() --- {} failed! See below:\n" \
                                   "{}"
                        msg_fortran = "\n".join(["! {}".format(x) for x in k.get_plain_text()])
                        msg = template.format(self.runnable.__class__.__name__, msg_fortran)
                        self.__logger.critical(msg)

                        #
                        # if isinstance(E, IOError):
                        #   printOpenFiles()
                        #
                        # self.__logger.exception("%s failed" % self.runnable.__class__.__name__)
                        # print "EXITING SO THAT YOU CAN SEE THE ERROR"
                        # self.manager.exit()
                        # raise


                    self.manager._finish(self)
                    self.runnable = None
                    self.flag_idle = True
                else:
                    misses += 1
                    if misses >= 34:
                        flag_sleep = True
                        misses = 0
            if flag_sleep:
                T = 0.1
                time.sleep(T)
                flag_sleep = False




#!/usr/bin/env python
"""
Retrieves and prints a table of molecular constants from the NIST Chemistry Web Book [NISTRef]

To do so, it uses web scraping to navigate through several pages and parse the desired information
from the book web pages.

It does not provide a way to list the molecules yet, but will give an error if the molecule is not
found in the NIST web book.

Example:

    print-nist.py OH

**Note** This script was designed to work with **diatomic molecules** and may not work with other
         molecules.

**Warning** The source material online was known to contain mistakes (such as an underscore instead
            of a minus signal to indicate a negative number). We have identified a few of these,
            and build some workarounds. However, we recommend a close look at the information parsed
            before use.

**Disclaimer** This script may stop working if the NIST people update the Chemistry Web Book.

References:

[NISTRef] http://webbook.nist.gov/chemistry/
"""

import pyfant
import tabulate
import a99
import logging
import argparse


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument('-u', "--unicode", action="store_true",
                        help='Unicode output (default is to contain only ASCII characters)')
    parser.add_argument('formula', type=str, help='NIST formula', nargs=1)
    args = parser.parse_args()

    data, header, title = pyfant.get_nist_webbook_constants(args.formula, args.unicode)
    out = "\n*** {} ***\n\n{}".format(title, tabulate.tabulate(data, header))

    if not args.unicode:
        # Just in case some unicode characters are left
        out = out.encode("ascii", "replace").decode()

    print(out)







import struct
import numpy as np
import os
import a99
from f311 import DataFile


__all__ = ["FileModBin", "ModRecord", "FileModTxt", "FileOpa", "FileMoo",
           "MooRecord"]


class ModRecord(a99.AttrsPart):
    """
    Single record from a99 atmospheric model file

    Note: while a infile:modeles may have several 1200-byte records stored in it,
    this class only stores one of these records, specified by "inum"  argument
    of load()

    Imitates the logic of reader_modeles.f90::read_modele().

    Attributes match reader_modeles.f90:modeles_* (minus the "modeles_" prefix)
    """

    attrs = ["ntot", "teff", "glog", "asalog", "asalalf", "nhe", "tit", "tiabs",
           "nh", "teta", "pe", "pg", "log_tau_ross"]
    less_attrs = ["teff", "glog", "asalog"]

    def __init__(self):
        a99.AttrsPart.__init__(self)
        self.ntot = None
        self.teff = None
        self.glog = None
        self.asalog = None
        self.asalalf = None
        self.nhe = None
        self.tit = None
        self.tiabs = None
        self.nh = None
        self.teta = None
        self.pe = None
        self.pg = None
        self.log_tau_ross = None

    def __repr__(self):
        return "/"+self.one_liner_str()+"/"


MOD_REC_SIZE = 1200


class FileModBin(DataFile):
  """
  PFANT Atmospheric Model (binary file)

  File may have several 1200-byte records.

  Files created by the innermarcs executable (usually named "modeles.mod")

  Imitates the logic of reader_modeles.f90::read_modele().

  Attributes match reader_modeles.f90:modeles_* (minus the "modeles_" prefix)
  """
  default_filename = "modeles.mod"
  attrs = ["records"]
  flag_txt = False

  def __init__(self):
    DataFile.__init__(self)
    self.records = None

  def __len__(self):
    if self.records is None:
      raise RuntimeError("File not loaded yet, len() is undefined")
    return len(self.records)

  def _do_load(self, filename):

    if a99.is_text_file(filename):
        raise RuntimeError("File must be binary")

    b = os.path.getsize(filename)

    if b < MOD_REC_SIZE:
        raise RuntimeError("File too small")

    num_rec = int(b/MOD_REC_SIZE)-1
    self.records = []
    with open(filename, "rb") as h:

      for inum in range(1, num_rec+1):
        pos = MOD_REC_SIZE*(inum-1)  # position of beginning of record requested
        h.seek(pos)
        x = h.read(MOD_REC_SIZE)
        rec = ModRecord()
        _decode_mod_record(x, rec, inum)
        self.records.append(rec)

  def _do_save_as(self, filename):
      """Saves to file."""

      with open(filename, "wb") as h:
          for i, rec in enumerate(self.records):
              _encode_mod_record(rec, h)
          h.write(struct.pack("<i", 9999))
          h.write(b"\x00"*1196)

  def init_default(self):
    raise RuntimeError("Not applicable")


class FileModTxt(DataFile):
    """
    MARCS Atmospheric Model (text file)

    Not to be confused with PFANT Atmospheric Model (binary file), as
    both share the same ".mod" extension

    http://marcs.astro.uu.se/

    This file contains a single atmospheric model for a given
    (teff, glog, asalog)

    """

    attrs = ["record"]

    def __init__(self):
        DataFile.__init__(self)
        self.record = None

    def __len__(self):
        return 1

    def _do_load(self, filename):
        if not a99.is_text_file(filename):
            raise RuntimeError("File must be a text file")
        r = ModRecord()
        with open(filename, "rb") as h:
            def _skip():
                h.readline()
            r.tit = h.readline().strip()
            r.tiabs = ""
            r.teff = float(struct.unpack("7s", h.readline()[:7])[0])
            _skip()  # flux row
            r.glog = np.log10(float(struct.unpack("12s", h.readline()[:12])[0]))
            _skip()  # microturbulence parameter
            _skip()  # mass
            r.asalog, r.asalalf = list(map(float, struct.unpack("6s 6s", h.readline()[:12])))
            _skip()  # "1 cm radius for plane-parallel models"
            _skip()  # "Luminosity"
            _skip()  # "convection parameters"

            # Reads hydrogen mass fraction to use later to calculate the nh vector
            # X, Y and Z are the mass fractions of H, He, and metals respectively (X+Y+Z=1)
            h_frac = float(h.readline()[:8])

            _skip()  # "Logarithmic chemical number abundances, H always 12.00"
            # reads He abundance to calculate "nhe"
            r.nhe = 10**(float(struct.unpack("7x 7s", h.readline()[:14])[0])-12)
            for i in range(9):  # skips 9 rows (file always has 92 abundances)
                _skip()
            n = r.ntot = int(struct.unpack("4s", h.readline()[:4])[0])
            _skip()  # "Model structure"
            _skip()  # header "k lgTauR  lgTau5    Depth     T        Pe          Pg         Prad       Pturb"

            r.teta, r.pe, r.pg, r.log_tau_ross = np.zeros(n), np.zeros(n), \
             np.zeros(n), np.zeros(n)

            # reads log(tau(Rosseland)), T, Pe, Pg
            for i in range(n):
                qwe = h.readline()
                r.log_tau_ross[i], t, r.pe[i], r.pg[i] = list(map(float,
                 struct.unpack("3x 6s 19x 8s 12s 12s", qwe[:60])))
                r.teta[i] = 5040./t

            # reads rhox to use in nh calculation
            _skip()  # hreader "k lgTauR    KappaRoss   Density   Mu      Vconv   Fconv/F      RHOX"
            rhox = np.zeros(n)
            for i in range(n):
                rhox[i] = float(h.readline()[60:72])

            # calculates nh using the formula in "transosmarcsok3.f"
            r.nh = rhox*(6.022142e23*h_frac)


        self.record = r


class FileOpa(DataFile):
    """MARCS ".opa" (opacity model) file format.

    Reference: http://marcs.astro.uu.se
    """

    default_filename = "modeles.opa"
    attrs = ["ndp", "swave", "nwav"]

    def __init__(self):
        DataFile.__init__(self)

        # # Global properties of the opacity model file
        # the 4-byte standard model code 'MRXF'
        self.mcode = None
        # number of depth points (=56)
        self.ndp = None
        # standard wavelength for the continuous optical depth (tau)
        # scale and for the total standard opacity (ops)
        self.swave = None
        # number of wavelengths for which continuous absorption and
        # scattering opacities are given. These are chosen so that
        # linear interpolation should suffice for any wavelength
        # (=1071)
        self.nwav = None
        # wav(j): wavelengths for which opacities are given
        self.wav = None

        # # Model structure
        # rad(k): radius, normalized on the outermost point, k=1. For use with
        # spherical radiative transfer. For plane-parallel models rad == 1.0.
        self.rad = None
        # tau(k): continuumm optical depth at the standard wavelength swave
        self.tau = None
        # t(k): temperature (K)
        self.t = None
        # pe(k): electron pressure (dyn/cm2)
        self.pe = None
        # pg(k)  = total gas pressure (dyn/cm2)
        self.pg = None
        # rho(k) = densigy (g/cm3)
        self.rho = None
        # xi(k)  = microturbulence parameter (km/s)
        self.xi = None
        # ops(k) = continuumm opacity at the standard wavelength (cm2/g)
        self.ops = None

        # # Wavelength-dependent opacities
        # abs(j,k) = specific continuous absorption opacity (cm2/g)
        self.abs = None
        # sca(j,k) = specific continuous scattering opacity (cm2/g)
        self.sca = None

        # logarithmic number abundances of the 92 first chemical
        # elements on a scale where the hydrogen abundance=12.00
        self.abund = None

    def cut(self, llzero, llfin):
        """Keeps only region within lambda interval [llzero, llfin]."""
        mask = np.logical_and(llzero <= self.wav, self.wav <= llfin)
        self.sca = self.sca[mask, :]
        self.abs = self.abs[mask, :]
        self.wav = self.wav[mask]
        self.nwav = len(self.wav)

    def _do_load(self, filename):
        """Loads from file.

        Based on http://marcs.astro.uu.se/documents/auxiliary/readopa.f
        """



        with open(filename, "rb") as h:
            self.mcode, self.ndp, self.swave = struct.unpack("1x 4s 5s 10s", h.readline().strip(b"\n"))
            if self.mcode != b"MRXF":
                # Does not satisfy magic string
                raise RuntimeError("Model code '{0!s}' is not 'MRXF'".format(self.mcode))
            self.ndp = int(self.ndp)
            self.swave = float(self.swave)

            self.nwav = int(h.readline())

            v, n_rows = a99.multirow_str_vector(h, self.nwav)
            self.wav = np.array(list(map(float, v)))

            self.rad, self.tau, self.t, self.pe, self.pg, self.rho, self.xi, \
            self.ops = np.zeros(self.ndp), np.zeros(self.ndp), np.zeros(self.ndp), \
                       np.zeros(self.ndp), np.zeros(self.ndp), np.zeros(self.ndp), \
                       np.zeros(self.ndp), np.zeros(self.ndp)
            self.abs = np.zeros((self.nwav, self.ndp))
            self.sca = np.zeros((self.nwav, self.ndp))
            for k in range(self.ndp):
                self.rad[k], self.tau[k], self.t[k], self.pe[k], self.pg[k], \
                self.rho[k], self.xi[k], self.ops[k] = a99.float_vector(h)
                v, n_rows = a99.multirow_str_vector(h, 2*self.nwav)
                abs_sca = np.array(list(map(float, v)))
                # This multiplication is performed as in original readopa.f
                self.abs[:, k] = abs_sca[0::2]*self.ops[k]
                self.sca[:, k] = abs_sca[1::2]*self.ops[k]

            v, n_rows = a99.multirow_str_vector(h, 92)
            self.abund = np.array(list(map(float, v)))



class MooRecord(a99.AttrsPart):
    """
    Single record from a ".moo" file

    The attributes are almost a concatenation of ModRecord with FileOpa
    attributes, except for redundant attributes, i.e.:
      - ntot = ndp (ntot kept)
      - pe, pg
    """

    attrs = ["ntot", "teff", "glog", "asalog", "asalalf", "nhe", "tit", "tiabs",
           "nh", "teta", "pe", "pg", "log_tau_ross", "swave", "nwav"]
    less_attrs = ["teff", "glog", "asalog"]

    def __init__(self):
        a99.AttrsPart.__init__(self)
        self.ntot = None
        self.teff = None
        self.glog = None
        self.asalog = None
        self.asalalf = None
        self.nhe = None
        self.tit = None
        self.tiabs = None
        self.nh = None
        self.teta = None
        self.pe = None
        self.pg = None
        self.log_tau_ross = None

        # These attributes exist in FileOpa as well, please check it for
        # descriptions.
        # Remaining attributes of FileOpa are not relevant
        self.swave = None
        self.nwav = None
        self.wav = None
        self.ops = None
        self.abs = None
        self.sca = None
        self.abund = None

    def __repr__(self):
        return "/"+self.one_liner_str()+"/"


    def from_marcs_files(self, file_mod, file_opa):
        """Copies attributes from a FileModTxt and a FileOpa."""

        aa = ["ntot", "teff", "glog", "asalog", "asalalf", "nhe", "tit",
              "tiabs", "nh", "teta", "pe", "pg", "log_tau_ross"]
        rec = file_mod.record
        for a in aa:
            self.__setattr__(a, rec.__getattribute__(a))
        aa = ["swave", "nwav", "wav", "ops", "abs", "sca", "abund"]
        for a in aa:
            self.__setattr__(a, file_opa.__getattribute__(a))



# Record size is record size in modeles.mod plus opacity part
# 13432 = 4+4+              (swave, nwav)
#         56*4+             (ops)
#         1071*(1+2*56)*4+  (wav, abs, sca)
#         92*4              (abund)
OPA_REC_SIZE = 484692
MOG_REC_SIZE = MOD_REC_SIZE+OPA_REC_SIZE
# Number of layers and wavelengths must have these fixed values because
# the ".mog" file I/O operations both in Python and Fortran are assuming thi.
# THe main reason for the assumption is that Fortran reads much faster when
# the READ() statement reads into whole array instead of implied DO.
#
# These fixed values seem to be part of the file specification anyway.
NTOT = 56    # must be 56
NWAV = 1071  # must be 1071

class FileMoo(DataFile):
    """
    Atmospheric model or grid of models (with opacities included)

    This file contains all the fields in modeles.mod, plus the opacity information
    """
    default_filename = "grid.moo"
    attrs = ["records"]
    flag_txt = False

    def __init__(self):
        DataFile.__init__(self)
        self.records = None

    def __len__(self):
        if self.records is None:
            raise RuntimeError("File not loaded yet, len() is undefined")
        return len(self.records)

    def _do_load(self, filename):

        if a99.is_text_file(filename):
            raise RuntimeError("File must be binary")

        b = os.path.getsize(filename)

        if b % MOG_REC_SIZE != 0:
            raise RuntimeError("Incorrect file size! Must be a multiple of {0:d}".format(
                               MOG_REC_SIZE))

        num_rec = int(b/MOG_REC_SIZE)
        self.records = []
        with open(filename, "rb") as h:
            for inum in range(1, num_rec+1):
                pos = MOG_REC_SIZE*(inum-1)  # position of beginning of record requested
                h.seek(pos)
                x = h.read(MOG_REC_SIZE)
                rec = MooRecord()
                _decode_mod_record(x, rec, inum)

                if rec.ntot != NTOT:
                    raise RuntimeError("Number of layers be {0:d}, not {1:d}".format(NTOT, rec.ntot))

                h.seek(pos+MOD_REC_SIZE)
                x = h.read(MOG_REC_SIZE)

                ostr = struct.Struct('<f i')
                [rec.swave,
                 rec.nwav] = ostr.unpack(x[:8])

                if rec.nwav != NWAV:
                    raise RuntimeError("Number of wavelengths must be {0:d}, not {1:d}".format(NWAV, rec.nwav))

                rec.ops = np.frombuffer(x, dtype='<f4', count=rec.ntot,
                                        offset=8)
                rec.wav = np.frombuffer(x, dtype='<f4', count=rec.nwav,
                                        offset=8+rec.ntot*4)
                rec.abs = np.frombuffer(x, dtype='<f4', count=rec.nwav*rec.ntot,
                 offset=8+(rec.ntot+rec.nwav)*4).reshape((rec.ntot, rec.nwav)).T
                rec.sca = np.frombuffer(x, dtype='<f4', count=rec.nwav*rec.ntot,
                 offset=8+(rec.ntot+rec.nwav*(1+rec.ntot))*4).reshape((rec.ntot, rec.nwav)).T
                rec.abund = np.frombuffer(x, dtype='<f4', count=92,
                 offset=8+(rec.ntot+rec.nwav*(1+2*rec.ntot))*4)
                self.records.append(rec)

    def _do_save_as(self, filename):
        """Saves to file."""

        with open(filename, "wb") as h:
            for i, rec in enumerate(self.records):
                assert rec.ntot == NTOT, \
                    "Number of layers be {0:d}, not {1:d}".format(NTOT, rec.ntot)
                assert rec.nwav == NWAV, \
                    "Number of wavelengths must be {0:d}, not {1:d}".format(NWAV, rec.nwav)

                _encode_mod_record(rec, h)

                ostr = struct.Struct('<f i')
                h.write(ostr.pack(rec.swave, rec.nwav))
                h.write(struct.pack("<"+"f"*rec.ntot, *rec.ops))
                s_temp ="<"+"f"*(rec.nwav*rec.ntot)


                h.write(struct.pack("<"+"f"*rec.nwav, *rec.wav))
                h.write(struct.pack(s_temp, *rec.abs.T.flatten()))
                h.write(struct.pack(s_temp, *rec.sca.T.flatten()))
                h.write(struct.pack("<"+"f"*92, *rec.abund))


    def init_default(self):
        raise RuntimeError("Not applicable")


_ostr = struct.Struct('<i 5f 20s 20s')  # header

def _decode_mod_record(x, rec, inum):
    """Decodes "x", binary record as in modeles.mod into "rec" attributes.
    Argument "inum" (record number) is for error reporting."""
    assert isinstance(rec, (ModRecord, MooRecord))
    [rec.ntot,
     rec.teff,
     rec.glog,
     rec.asalog,
     rec.asalalf,
     rec.nhe,
     rec.tit,
     rec.tiabs] = _ostr.unpack(x[:64])

    # This routine will read almost any binary, so we perform some range checks
    if not (1 < rec.ntot <= 1000):
        raise RuntimeError("record #{0:d}: ntot invalid: {1:d}".format(inum, rec.ntot))
    if not (100 < rec.teff < 100000):
        raise RuntimeError("record #{0:d}: teff invalid {1:g}".format(inum, rec.teff))

    v = np.frombuffer(x, dtype='<f4', count=rec.ntot*5, offset=64)
    w = np.reshape(v, (rec.ntot, 5))

    rec.nh, rec.teta, rec.pe, rec.pg, rec.log_tau_ross = [w[:, i] for i in range(5)]

def _encode_mod_record(rec, h):
    """Encodes record into open file."""
    assert isinstance(rec, (ModRecord, MooRecord))

    tit = rec.tit if isinstance(rec.tit, bytes) else rec.tit.encode("ascii")
    tiabs = rec.tit if isinstance(rec.tiabs, bytes) else rec.tiabs.encode("ascii")

    s = _ostr.pack(rec.ntot, rec.teff, rec.glog, rec.asalog,
                       rec.asalalf, rec.nhe, tit, tiabs)
    h.write(s)
    ny = 5 * rec.ntot
    y = np.reshape(np.vstack([rec.nh, rec.teta, rec.pe, rec.pg, rec.log_tau_ross]).T,
                   ny)
    h.write(struct.pack("<" + "f" * ny, *y))
    h.write(b"\x00" * (MOD_REC_SIZE-ny*4-64))  # fills record with \x0 to have 1200 bytes




import matplotlib.pyplot as plt, numpy, a99, numpy as np

__all__ = ["plot_molhistogram"]

def plot_molhistogram(mol, num_bins=100, color=None, label=None):
    """Plots histogram that gives a general profile of where the lines are wavelength-wise"""
    import pyfant
    assert isinstance(mol, pyfant.Molecule)

    x0 = mol.llzero
    x1 = mol.llfin
    data = mol.lmbdam

    bins = np.logspace(np.log10(x0), np.log10(x1), num_bins+1, endpoint=True)
    h = np.histogram(data, bins=bins, density=False)[0]
    widths = np.diff(bins)

    ax = plt.gca()
    plt.bar(bins[:-1], h, width=widths, align="edge", color=color, label=label)
    plt.xscale("log")

    plt.xlabel("Wavelength ($\AA$)")
    plt.ylabel("Number of lines")
    a99.format_BLB()



import subprocess
from .conf import *
import os
import a99
from collections import OrderedDict
import pyfant


__all__ = ["Runnable", "RunnableStatus", "ExecutableStatus", "PFANTExecutable",
           "Innewmarcs", "Hydro2", "Pfant", "Nulbad", "Combo"]


@a99.froze_it
class RunnableStatus(object):

    def __init__(self, runnable):
        assert isinstance(runnable, Runnable)
        self.runnable = runnable

    def __str__(self):
        l = []
        if self.runnable.flag_finished:
            l.append("finished")
        if self.runnable.flag_running:
            l.append("running")
        if self.runnable.flag_killed:
            l.append("*killed*")
        if self.runnable.flag_error:
            l.append("*error*")
        if self.runnable.error_message:
            l.append("*" + str(self.runnable.error_message) + "*")
        if len(l) > 0:
            return " ".join(l)
        return "?"

@a99.froze_it
class ExecutableStatus(object):
    """Stores status related to PFANTExecutable for reporting purposes."""
    
    def __init__(self, executable):
        assert isinstance(executable, PFANTExecutable)
        self.exe_filename = executable.__class__.__name__.lower()
        self.executable = executable
        # used by pfant only
        self.ikey = None
        # used by pfant only
        self.ikeytot = None

    def __str__(self):
        l = []
        if self.exe_filename is not None:
            l.append(self.exe_filename)
        if self.executable.flag_finished:
            l.append("finished")
        if self.executable.flag_running:
            l.append("running")
        if self.executable.flag_killed:
            l.append("*killed*")
        if self.executable.flag_error:
            l.append("*error*")
        if self.executable.error_message:
            l.append("*" + str(self.executable.error_message) + "*")
        if self.ikey is not None:
            l.append("%5.1f %% (%d/%d)" % (100.*self.ikey/self.ikeytot, self.ikey, self.ikeytot))
        if self.executable.returncode is not None:
            l.append("returncode=%d" % self.executable.returncode)
        if len(l) > 0:
            return " ".join(l)
        return "?"


class Runnable(object):
    """
    Object with a run() method.

    The run() method is blocking, i.e., it only returns when running is done.

    This is a base class for PFANTExecutable and Combo.
    """

    @property
    def flag_finished(self):
        return self._flag_finished
    @property
    def flag_running(self):
        return self._flag_running
    @property
    def flag_killed(self):
        return self._flag_killed
    @property
    def flag_error(self):
        return self._flag_error
    @property
    def error_message(self):
        return self._error_message
    @property
    def flag_success(self):
        return self._flag_finished and not self._flag_error and not self._flag_killed

    @property
    def sid(self):
        return self._get_sid()

    @property
    def result(self):
        return self._result

    def __init__(self):
        self.name = a99.random_name()
        # Is running?
        self._flag_running = False
        # Is finished?
        self._flag_finished = False
        # Was killed?
        self._flag_killed = False
        # Had error?
        self._flag_error = False
        # Will contain error message if finished with error
        self._error_message = ""
        # Will contain results DataFile objects (keys vary depending on the Runnable subclass
        self._result = {}

    def get_status(self):
        raise NotImplementedError()

    def run(self):
        raise NotImplementedError()

    def clean(self):
        """Deletes session directory"""
        self.sid.clean()

    def kill(self):
        raise NotImplementedError()

    def load_result(self):
        """Abstract. Override this method to open the result file(s) particular to the
        executable.

        Methods in subclasses should populate self._result
        """

    def reset(self):
        """Prepares to run again."""
        assert not self._flag_running, "Cannot reset if running"
        self._flag_finished = False
        self._flag_killed = False
        self._flag_error = False
        self._error_message = ""
        if self.sid.id:
            self.sid.clean(False)

    def _get_sid(self):
        raise NotImplementedError()


class PFANTExecutable(Runnable):
    """
    PFANT executables common ancestor class, or any other program
    """

    # Set at descendant class with a ftpyfant.conf.FOR_* value
    sequence_index = -1

    @property
    def returncode(self):
        return self.__returncode

    @property
    def exe_path(self):
        return self._exe_path

    @exe_path.setter
    def exe_path(self, x):
        self._exe_path = x

    @property
    def conf(self):
        return self.__conf

    @conf.setter
    def conf(self, x):
        self.__conf = x

    @property
    def opt(self):
        """Wraps self.__conf.opt"""
        return self.__conf.opt

    @opt.setter
    def opt(self, x):
        """Wraps self.__conf.opt"""
        self.__conf.opt = x

    def __init__(self):
        Runnable.__init__(self)
        # # Protected variables
        # Full path to executable (including executable name)
        self._exe_path = "none"
        # ExecutableStatus instance
        self._status = ExecutableStatus(self)

        # # Private variables
        # fill be filled with self.popen.returncode in due time
        self.__returncode = None
        # Created by _run()
        self.__popen = None
        # Conf instance
        self.__conf = Conf()

    def run(self):
        """Runs executable.

        Blocking routine. Only returns when executable finishes running.
        """
        assert not self._flag_running, "Already running"
        assert not self._flag_finished, "Already finished"
        self.conf.configure([self.sequence_index])
        try:
            self.conf.logger.debug("Running %s '%s'" % (self.__class__.__name__.lower(), self.name))
            self.__run()
        finally:
            self.conf.close_popen_text_dest()

    def clean(self):
        """Deletes session directory"""
        self.sid.clean()

    def run_from_combo(self):
        """Alternative to run executable (called from Combo class).

        This routine bypasses all the configuration that is done prior to running.
        (Combo.configure() will do the necessary configuration).
        """
        assert not self._flag_running, "Already running"
        assert not self._flag_finished, "Already finished"
        self.__run()

    def kill(self):
        self._flag_killed = True
        if self._flag_running:
            self.__popen.kill()

    def get_status(self):
        return self._status

    def _get_sid(self):
        return self.__conf.sid

    def __run(self):
        """Called both from run() and run_from_combo()."""

        self.conf.logger.debug("FLPREFIX NOW "+str(self.conf.opt.flprefix))

        args = self.conf.get_args()
        cmd_words = [self._exe_path] + args

        s = " ".join(cmd_words)
        self.conf.logger.debug(s)
        # logs command-line to file and closes it.
        with open(self.conf.sid.join_with_session_dir("commands.log"), "a") as h:
            h.write(s+"\n\n")
        emsg = ""
        try:
            self.__popen = subprocess.Popen(cmd_words, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            self._flag_running = True
            try:
                if self.conf.popen_text_dest is not None:
                    for line in self.__popen.stdout:
                        # In Python 3, line is bytes, write() cannot deal with that
                        self.conf.popen_text_dest.write(line.decode("ascii"))
            finally:
                self.__popen.stdout.close()

            # blocks execution until finished
            self.__popen.wait()
            self.__popen.poll()

            if self.__popen.returncode != 0:
                raise pyfant.FailedError("%s failed (returncode=%s)" % (self.__class__.__name__.lower(), self.__popen.returncode))

        except Exception as e:
            flag_re_raise = True
            if self.__popen:
                self.__popen.poll()
                if isinstance(e, pyfant.FailedError) and self._flag_killed:
                    # dismisses error if explicitly killed
                    flag_re_raise = False
                elif isinstance(e, IOError) and self.__popen.returncode == 0:
                    # Sometimes a IOError is raised even if Fortran executes
                    # successfully, so the error is dismissed
                    self.conf.logger.warning("Harmless error in: %s %s" %
                     (self.conf.sid.dir, self.get_status()))
                    self.conf.logger.warning(str(e))
                    flag_re_raise = False

            if flag_re_raise:
                self._error_message = e.__class__.__name__+": "+str(e)
                self._flag_error = True
                raise
        finally:
            self._flag_finished = True
            self._flag_running = False
            if self.__popen is not None:
                self.__returncode = self.__popen.returncode
            self.conf.logger.debug(str(self._status))


@a99.froze_it
class Innewmarcs(PFANTExecutable):
    """Class representing the innewmarcs executable."""

    sequence_index = FOR_INNEWMARCS

    def __init__(self):
        PFANTExecutable.__init__(self)
        self._exe_path = "innewmarcs"

        # FileModBin object
        self.modeles = None

    def load_result(self):
        file_mod = pyfant.FileModBin()
        filepath = self.conf.get_fn_modeles()
        file_mod.load(filepath)
        # Assigns .modeles for backward compatibility
        self.modeles = self._result["modeles"] = file_mod


@a99.froze_it
class Hydro2(PFANTExecutable):
    """Class representing the hydro2 executable."""

    sequence_index = FOR_HYDRO2

    def __init__(self):
        PFANTExecutable.__init__(self)
        self._exe_path = "hydro2"

    def load_result(self):
        """
        Makes self._result["profiles"] = {filename0: FileToH0, filename1: FileToH1, ...}

        Tolerant to non-existing files, but will crash if existing files pointed to at the
        FileHmap fail to load.
        """

        fhmap = self.conf.get_file_hmap()
        if fhmap is None:
            raise RuntimeError("Cannot find my hmap file")

        profiles = self._result["profiles"] = OrderedDict()
        for row in fhmap.rows:
            value = None
            if os.path.exists(row.fn):
                fh = pyfant.FileToH()
                fh.load(row.fn)
                value = fh

            profiles[row.fn] = value



@a99.froze_it
class Pfant(PFANTExecutable):
    """Class representing the pfant executable."""

    sequence_index = FOR_PFANT

    def __init__(self):
        PFANTExecutable.__init__(self)
        self._exe_path = "pfant"  # Path to PFANT executable (including executable name)

        # ** Variables assigned by poll_progress()
        self.ikey = None  # Current iteration (Fortran: ikey)
        self.ikeytot = None  # Current iteration (Fortran: ikeytot)

        # # Spectra loaded after pfant runs if self.flag_open_result is True

        # .spec file, e.g., flux.spec
        self.spec = None
        # .cont file, e.g., flux.cont
        self.cont = None
        # .norm file, e.g., flux.norm
        self.norm = None

    def get_status(self):
        """
        Tries to open progress indicator file. If succeeds, stores
        current iteration in attribute ikey, and
        number of iterations in attribute ikeytot.

        """
        p = self.conf.opt.fn_progress or "progress.txt"
        ret = self._status
        if (not self.ikey or self.ikey < self.ikeytot) and os.path.isfile(p):
            with open(p) as h:
                try:
                    t = list(map(int, h.readline().split("/")))
                    ret.ikey = t[0]
                    ret.ikeytot = t[1]
                except ValueError:
                    # Ignores conversion errors
                    pass
        return ret

    def load_result(self):

        for type_ in ("norm", "cont", "spec"):
            filepath = self.conf.get_pfant_output_filepath(type_)
            file_sp = pyfant.FileSpectrumPfant()
            file_sp.load(filepath)
            self.__setattr__(type_, file_sp.spectrum)
            self._result[type_] = file_sp.spectrum


@a99.froze_it
class Nulbad(PFANTExecutable):
    """Class representing the nulbad executable."""

    sequence_index = FOR_NULBAD

    def __init__(self):
        PFANTExecutable.__init__(self)
        self._exe_path = "nulbad"

        # nulbad output
        self.convolved = None

    def load_result(self):
        file_sp = pyfant.FileSpectrumNulbad()
        filepath = self.conf.get_nulbad_output_filepath()
        try:
            file_sp.load(filepath)
        except:
            print("WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW")
            raise
        self.convolved = file_sp.spectrum
        self._result["convolved"] = file_sp.spectrum


@a99.froze_it

class Combo(Runnable):
    """
    Runs sequence of executables: innermarcs, hydro2, pfant, nulbad.

    Args:
      sequence (optional): sequence of executables to run. Defaults to
        [e_innewmarcs, e_hydro2, e_pfant, e_nulbad]. If you want to run only
        pfant and nulbad, for example, you can pass [e_pfant, e_nulbad]

    There are several restrictions imposed
    - files are created inside a session directory such as session123456
    - all four executables must be in the same directory
    """

    @property
    def running_exe(self):
        """Returns the current or last running exe."""
        return self.__running_exe

    @property
    def exe_dir(self):
        return self.__exe_dir
    @exe_dir.setter
    def exe_dir(self, x):
        self.__exe_dir = x

    @property
    def sequence(self):
        return self.__sequence
    @sequence.setter
    def sequence(self, x):
          self.__sequence = x

    @property
    def innewmarcs(self):
        return self.__innewmarcs

    @property
    def hydro2(self):
        return self.__hydro2

    @property
    def pfant(self):
        return self.__pfant

    @property
    def nulbad(self):
        return self.__nulbad

    @property
    def conf(self):
        return self.__conf

    @conf.setter
    def conf(self, x):
        self.__conf = x

    def __init__(self, sequence=None):
        Runnable.__init__(self)
        # # Configuration
        # Directory containing the 4 executables
        self.__exe_dir = ""
        # Whether to display Fortran messages in the terminal.
        # Note: a file named "<session dir>/fortran.log" is always created.
        self.__flag_log_console = True

        # Executables to run
        # order is irrelevant (will be sorted anyway).
        self.__sequence = [FOR_INNEWMARCS, FOR_HYDRO2, FOR_PFANT, FOR_NULBAD] \
            if sequence is None else translate_sequence(sequence)

        # ** PFANTExecutable instances
        self.__innewmarcs = Innewmarcs()
        self.__hydro2 = Hydro2()
        self.__pfant = Pfant()
        self.__nulbad = Nulbad()

        # ** Internal variables
        self.__running_exe = None  # PFANTExecutable object currently running
        # ComboStatus instance
        self.__status = RunnableStatus(self)
        # Conf instance
        self.__conf = Conf()

    def get_exes(self):
        """Returns exe objects in a list according with self.sequence."""

        map = [(FOR_INNEWMARCS, self.__innewmarcs), (FOR_HYDRO2, self.__hydro2), (FOR_PFANT, self.__pfant),
               (FOR_NULBAD, self.__nulbad)]
        res = []
        ii, ee = list(zip(*map))
        self.__sequence.sort()
        for i_exe in self.__sequence:
            if i_exe in ii:
                res.append(ee[ii.index(i_exe)])
        return res

    def run(self):
        assert not self._flag_running, "Already running"
        assert not self._flag_finished, "Already finished"
        self._flag_running = True
        try:
            self.conf.configure(self.__sequence)
            self.conf.logger.debug("Running %s '%s'" % (self.__class__.__name__.lower(), self.name))
            for e in self.get_exes():
                self.__running_exe = e
                e.conf = self.conf
                e.run_from_combo()
        except Exception as e:
            self._flag_error = True
            self._error_message = "Combo: "+str(e)
            raise
        finally:
            self.conf.close_popen_text_dest()
            self._flag_running = False
            self._flag_finished = True

    def load_result(self):
        """Calls load_result() for all contained executables, then collects all into self._result"""
        for exe in self.get_exes():
            exe.load_result()

            # Bug checking: Executables must not repeat the names of their results,
            # so that they can be all collected in Combo._result
            for key in exe._result:
                if key in self._result:
                    raise AssertionError("Key '{}' is in {} but was seen previously by Combo".
                                         format(key, exe.__class__))
            self._result.update(exe._result)

    def kill(self):
        self._flag_killed = True
        if self._flag_running:
            if self.__running_exe:
                self.__running_exe.kill()

    def get_status(self):
        """Returns status of running executable or self._status."""
        if self.__running_exe:
            return self.__running_exe.get_status()
        else:
            return self.__status

    def reset(self):
        Runnable.reset(self)
        ee = self.get_exes()
        for e in ee:
            e.reset()

    def _get_sid(self):
        return self.__conf.sid



"""Widget to edit a series of points on top of a spectrum"""

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *
import a99
import matplotlib.pyplot as plt
import copy
import f311


__all__ = ["WContinua"]

_COLOR_LINE = (.91, .13, 0)
_COLOR_MARKER = (.92, .132, 0)
_COLOR_SPECTRUM = (0., .2, .9)
_LINESTYLE_CONTINUUM = "solid"
_LINESTYLE_SPECTRUM = "dashed"
_LINEWIDTH = 1
_MARKER = "s"


class WContinua(a99.WConfigEditor):

    @property
    def continua(self):
        return self._continua

    @continua.setter
    def continua(self, value):
        self._continua = value if value is not None else []

    @property
    def spectrum(self):
        return self._spectrum

    @spectrum.setter
    def spectrum(self, value):
        self._spectrum = value
        self._update_self()
        self._draw()

    def __init__(self, parent_form):
        a99.WConfigEditor.__init__(self, parent_form)

        self._spectrum = None
        # matplotlib plot corresponding to self._spectrum
        self._plot_sp = None
        self._draw_last_name = "!&*@*@*#@#*@#"  # improbable name
        self._continua = []

        lw0 = self.layout_main = QVBoxLayout()
        self.setLayout(lw0)

        sp = self.splitter = QSplitter(Qt.Horizontal)
        self.layout_editor.addWidget(sp)

        wsp = a99.keep_ref(QWidget())
        sp.addWidget(wsp)
        ly_wsp = QVBoxLayout(wsp)

        toolbar0 = self._get_toolbar0()
        ly_wsp.addWidget(toolbar0)

        a = self.tableWidget = self._get_tableWidget0()
        ly_wsp.addWidget(a)

        # Matplotlib area

        wm = a99.keep_ref(QWidget())
        # a99.set_margin(wm, 0)
        sp.addWidget(wm)
        self.figure, self.canvas, self.lfig = a99.get_matplotlib_layout(wm)
        self.canvas.mpl_connect('button_press_event', self._on_plot_click)

        # name, property
        self._map = [
            a99.CEMapItem("continua", self),
        ]

    # GUI artists

    def _get_toolbar0(self):
        # ## Actions
        action = self.action_insert = QAction(a99.get_icon("list-add"), "&Insert...", self)
        action.triggered.connect(self.on_insert)
        action.setShortcut("Ins")

        action = self.action_edit = QAction(a99.get_icon("gtk-edit"), "&Edit...", self)
        action.triggered.connect(self.on_edit)
        action.setShortcut("Enter")

        action = self.action_delete = QAction(a99.get_icon("trash"), "&Delete", self)
        action.setShortcut("Del")
        action.triggered.connect(self.on_delete)

        # Adds actions to toolbar

        tb = QToolBar()
        tb.setOrientation(Qt.Horizontal)
        tb.addAction(self.action_insert)
        tb.addAction(self.action_edit)
        tb.addAction(self.action_delete)

        return tb

    def _get_tableWidget0(self):
        a = QTableWidget()
        a.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setAlternatingRowColors(True)
        a.setEditTriggers(QAbstractItemView.NoEditTriggers)
        a.setFont(a99.MONO_FONT)
        # a.installEventFilter(self)
        # a.setContextMenuPolicy(Qt.CustomContextMenu)
        # a.customContextMenuRequested.connect(self.on_twSpectra_customContextMenuRequested)
        # a.setSortingEnabled(True)
        # a.cellChanged.connect(self.on_twSpectra_cellChanged)
        a.itemSelectionChanged.connect(self._on_tableWidget_itemSelectionChanged)
        return a

    # WConfigEditor override

    def _after_update_gui(self):
        self._populate()
        self._draw()

    def _after_update_fobj(self):
        newcontinua = [copy.copy(continuum) for continuum in self._f.obj["continua"]]
        for continuum in newcontinua:
            continuum["plot0"] = None
            continuum["plot1"] = None

        self._f.obj["continua"] = newcontinua

    # Slots

    def _on_plot_click(self, event):
        print('{} click: button={}, x={}, y={}, xdata={}, ydata={}'.format(
              'double' if event.dblclick else 'single', event.button,
               event.x, event.y, event.xdata, event.ydata))

        x, y = event.xdata, event.ydata

        if x is None:
            return

        sp = self.spectrum
        if sp is not None:
            idx = a99.index_nearest(sp.x, x)
            print("Lambda is actually {}".format(sp.x[idx]))

        if event.button == 1:
            self._insert_point(x, y)
        elif event.button == 3:
            self._delete_nearest(x, y)

    def _on_tableWidget_itemSelectionChanged(self):
        self._update_fobj()
        self._draw()
        self.changed.emit()

    def _sth_changed(self):
        self._update_fobj()
        self.changed.emit()

    def on_insert(self):
        if self._do_on_insert():
            self._update_fobj()
            self.changed.emit()

    def on_edit(self):
        if self._get_continuum_index() is None:
            return
        if self._do_on_edit():
            self._update_fobj()
            self.changed.emit()

    def on_delete(self):
        if self._get_continuum_index() is None:
            return
        if self._do_on_delete():
            self._update_fobj()
            self.changed.emit()

    # Internal

    def _get_new_continuum(self, name):
        return {"name": name, "points": [], "plot0": None, "plot1": None}

    def _get_continuum_index(self):
        """Returns index of selected continuum"""
        n = len(self._continua)
        if n == 0:
            return None
        index = self.tableWidget.currentRow()
        if index > len(self._continua)-1:
            return None
        return index

    def _get_continuum(self):
        """Returns selected continuum"""
        index = self._get_continuum_index()
        if index is None:
            return None
        return self._continua[index]

    def _get_continuum_name(self):
        """Returns selected continuum"""
        cuum = self._get_continuum()
        return None if cuum is None else cuum["name"]

    def _get_points(self):
        return self._get_continuum()["points"]

    def _do_on_insert(self):
        name = a99.random_name()
        while True:
            params = a99.Parameters({"name": name})
            form = a99.XParametersEditor(specs=params, title="Insert continuum")
            r = form.exec_()
            if not (r == QDialog.Accepted):
                return

            name = form.get_kwargs()["name"]
            if self._find_continuum_name(name) is not None:
                a99.show_error("Name already exists")
                continue

            self._continua.append(self._get_new_continuum(name))
            continuum = self._get_continuum()
            self._populate()
            if continuum is not None:
                self._position_at_name(continuum["name"])

            return True

    def _do_on_edit(self):
        continuum = self._get_continuum()
        name = continuum["name"]
        while True:
            r, form = a99.show_edit_form({"name": name}, None, "Edit continuum name")
            if not (r == QDialog.Accepted):
                return

            newname = form.get_kwargs()["name"]
            if any([cuum_["name"] == newname for cuum_ in self._continua if cuum_ != continuum]):
                a99.show_error("Name already exists")
                continue

            continuum["name"] = newname
            self.tableWidget.item(self._get_continuum_index(), 0).setText(newname)

            return True

    def _do_on_delete(self):
        r = QMessageBox.question(None, "Delete continuum",
                                 "Are you sure?",
                                 QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if r == QMessageBox.Yes:
            index = self._get_continuum_index()
            del self._continua[index]
            self._populate()
            if index <= len(self._continua):
                self.tableWidget.setCurrentCell(index, 0)

            return True


    def _position_at_name(self, name):
        """Positions table widget at continuum given by name"""
        a = self.tableWidget
        idx = self._find_continuum_name(name)
        save = a.blockSignals(True)
        a.setCurrentCell(idx, 0)
        a.blockSignals(save)

    def _find_continuum_name(self, name):
        """Returns index of continuum or None"""
        a = self.tableWidget
        for i in range(a.rowCount()):
            if a.item(i, 0).text() == name:
                return i
        return None

    def _update_self(self):
        """Changes to internal state sensible to current internal state"""






# TODO Auto make continuum to start with
        # sp = self._spectrum
        # if self._points is None and sp is not None:
        #     if False:
        #         y0, y1 = sp.y[0], sp.y[-1]
        #     else:
        #         y0, y1 = 1., 1.
        #     self._points = [(sp.l0, y0), (sp.lf, y1)]


    def _lose_plots(self):

        for cuum in self._continua:
            cuum["plot0"] = None
            cuum["plot1"] = None
        self._plot_sp = None

    def _draw(self):

        def _ensure_new_fig():
            fig = self.figure
            fig.clear()
            plt.figure(fig.number)
            while True:
                yield
        enfig = _ensure_new_fig()

        try:
            a99.format_BLB()

            name = self._get_continuum_name()
            if name != self._draw_last_name:
                self._lose_plots()
                next(enfig)
            self._draw_last_name = name


            # Spectrum
            sp = self._spectrum
            if sp is not None:
                ######################### next(enfig)
                if self._plot_sp is None:
                    self._plot_sp, = plt.plot(sp.x, sp.y, label=str(sp.title), color=_COLOR_SPECTRUM,
                             linestyle=_LINESTYLE_SPECTRUM, linewidth=_LINEWIDTH)
                    setup = f311.PlotSpectrumSetup()
                    plt.xlabel(setup.fmt_xlabel.format(sp))
                else:
                    self._update_plot_data(self._plot_sp, zip(sp.x, sp.y))

                xmin, xmax, ymin, ymax, xspan, yspan = f311.calc_max_min([sp])
                ymin, ymax = min(0., ymin), max(1., ymax)

                _T = 0.02  # fraction of extra space on left, right, top, bottom of graphics
                plt.xlim([xmin - xspan * _T, xmax + xspan * _T])
                plt.ylim([ymin - yspan * _T, ymax + yspan * _T])


            # Continua
            cuum = self._get_continuum()
            if cuum is not None:
                points = cuum["points"]

                p0 = cuum["plot0"]
                if p0:
                    self._update_plot_data(p0, points)
                    self._update_plot_data(cuum["plot1"], points)
                else:
                    xx, yy = self._get_xx_yy(points)
                    ########################################## next(enfig)
                    cuum["plot0"], = plt.plot(xx, yy, c=_COLOR_LINE, linestyle=_LINESTYLE_CONTINUUM,
                         linewidth=_LINEWIDTH)
                    cuum["plot1"], = plt.plot(xx, yy, c=_COLOR_MARKER, linestyle="None", marker=_MARKER)

            try:
                self.figure.tight_layout()
            except:
                a99.get_python_logger().exception("WContinua._draw()")

        except:
            a99.get_python_logger().exception("WContinua._draw()")
        finally:
            self.canvas.draw()

    def _update_plot_data(self, plot, points):
        xx, yy = self._get_xx_yy(points)
        plot.set_xdata(xx)
        plot.set_ydata(yy)

    def _get_xx_yy(self, points):
        try:
            xx, yy = zip(*points)
        except (ValueError, TypeError):
            xx, yy = [], []
        return xx, yy

    def _insert_point(self, x, y):
        cuum = self._get_continuum()
        if cuum is None:
            return
        pp = cuum["points"]
        p = (x, y)
        idx = -1
        if len(pp) > 0:
            xx, _ = zip(*pp)
            idx = a99.BSearchCeil(xx, x)
        if idx == -1:
            pp.append(p)
        else:
            pp.insert(idx, p)

        self._draw()
        self.changed.emit()

    def _delete_nearest(self, x, y):
        """Deletes point nearest to (x, y) (actually ignores the y)"""

        cuum = self._get_continuum()
        if cuum is None:
            return
        pp = cuum["points"]

        if len(pp) == 0:
            return

        xx, _ = zip(*pp)
        index = a99.index_nearest(xx, x)
        del pp[index]

        self._draw()
        self.changed.emit()

    def _populate(self):
        _HEADERS = ["Name"]

        cc = self._continua
        nr, nc = len(cc), len(_HEADERS)
        t = self.tableWidget
        a99.reset_table_widget(t, nr, nc)
        t.setHorizontalHeaderLabels(_HEADERS)
        i_row = 0
        for i, continuum in enumerate(cc):
            item = QTableWidgetItem(continuum["name"])
            t.setItem(i_row, 0, item)

            i_row += 1

        t.resizeColumnsToContents()



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from .a_WDBMolecule import WDBMolecule
from .a_WDBState import WDBState
from .a_WDBSystemPfantmolFCF import *
import os


__all__ = ["WFileMolDB"]


class WFileMolDB(a99.WEditor):


    def __init__(self, *args):
        a99.WEditor.__init__(self, *args)

        # # Main layout & splitter
        lmain = self.layout_editor

        sp = self.splitter_bidon = QSplitter(Qt.Horizontal)

        # ## First widget of splitter: molecules
        w0 = self.keep_ref(QWidget())
        l0 = QVBoxLayout(w0)
        a99.set_margin(l0, 2)
        l0.setSpacing(2)

        a = self.title_mol = QLabel(a99.format_title0("Molecules (Alt+&M)"))
        l0.addWidget(a)

        w = self.keep_ref(QLineEdit())
        l0.addWidget(w)

        w = self.w_mol = WDBMolecule(self.parent_form)
        # **Note** Cannot set buddy to w itself because it has a Qt.NoFocus policy
        a.setBuddy(w.tableWidget)
        w.id_changed.connect(self.mol_id_changed)
        w.changed.connect(self.changed)
        l0.addWidget(w)

        # ## Second widget of splitter: tab widget containing the rest
        w1 = self.tabWidget = QTabWidget(self)

        # ### First tab: systems, PFANT molecules, and Franck-Condon factors
        w = self.w_system = WDBSystemPFANTMolFCF(self.parent_form)
        w.changed.connect(self.changed)
        w1.addTab(self.w_system, "Electronic systems (Alt+&E)")

        # ### Second tab: NIST Chemistry Web Book data
        w = self.w_state = WDBState(self.parent_form)
        w.changed.connect(self.changed)
        w1.addTab(self.w_state, "States from NIST (Alt+&S)")

        sp.addWidget(w0)
        sp.addWidget(w1)
        # sp.setStretchFactor(0, 1)
        # sp.setStretchFactor(1, 2)

        lmain.addWidget(sp)

        # # Final adjustments
        a99.nerdify(self)

    def load(self, fobj):
        a99.WEditor.load(self, fobj)
        self.w_mol._move_to_first()

    def _do_load(self, fobj):
        self._f = fobj
        self.w_mol.f = fobj
        self.w_system.f = fobj
        self.w_state.f = fobj

    def mol_id_changed(self):
        id_ = self.w_mol.id
        row = self.w_mol.row
        self.w_system.set_id_molecule(id_)
        self.w_state.set_id_molecule(id_)





__all__ = ["FileAbonds"]

import struct
import a99
from f311 import DataFile
from ..basic import adjust_atomic_symbol
import re
from .filedissoc import FileDissoc
import tabulate
import pyfant

class FileAbonds(DataFile):
    """PFANT Stellar Chemical Abundances"""

    default_filename = "abonds.dat"
    attrs = ["notes", "ele", "abol", "notes_per_ele"]
    editors = ["abed.py", "x.py"]

    def __init__(self):
        DataFile.__init__(self)
        # list of atomic symbols
        self.ele = []
        # corresponding abundances
        self.abol = []
        # notes per element, ignored by pfant
        self.notes_per_ele = []
        # overall
        self.notes = ""

    def __str__(self):
        data = zip(self.ele, self.abol, self.notes_per_ele)
        headers = ["El", "Abund", "Notes"]

        notes = ""
        if self.notes:
            notes = "\n\nNotes\n-----\n{}".format(self.notes)

        return tabulate.tabulate(data, headers)+notes

    def __add__(self, other):
        if not isinstance(other, FileAbonds):
            raise TypeError("I don't want a {}".format(other.__class__.__name__))
        return "I dont know"

    def __len__(self):
        """Returns length of "ele" attribute."""
        return len(self.ele)

    def __getitem__(self, ele):
        """Return abundance given element symbol.

        notes_per_ele may be None if index is beyond len.
        """
        import pyfant as pf
        i = self.ele.index(pf.adjust_atomic_symbol(ele))
        return self.abol[i]

    def __setitem__(self, ele, abol):
        import pyfant as pf
        i = self.ele.index(pf.adjust_atomic_symbol(ele))
        self.abol[i] = abol

    def _do_load(self, filename):
        self.abol, self.ele, self.notes_per_ele = [], [], []

        ostr = struct.Struct("1x 2s 6s")
        with open(filename, "r") as h:
            for s in h:
                if len(s) > 0:
                    if s[0] == "1":  # sign to stop reading file
                        if len(self) == 0:
                            # We need at least one element in order increase the amount of
                            # file validation
                            raise RuntimeError("'EOF' marker found at beginning of file, I need at least one element")

                        self.notes = s[10:].replace("<br>", "\n")
                        break
                [ele, abol, notes] = s[1:3], s[3:9], s[10:]

                if not re.search(r'[a-z]', ele, re.IGNORECASE):
                    raise RuntimeError("Invalid element symbol: '%s'" % ele.strip())

                self.ele.append(adjust_atomic_symbol(ele))
                self.abol.append(float(abol))
                self.notes_per_ele.append(notes.strip())

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            h.writelines([' %-2s%6.2f %s\n' % (self.ele[i], self.abol[i], self.notes_per_ele[i])
                          for i in range(len(self))])
            c = (" "*9)+self.notes.replace("\n", "<br>") if len(self.notes) > 0 else ""
            h.writelines(['1'+c+'\n', '1\n'])

    def get_file_dissoc(self):
        """Creates a new FileDissoc object.

        To do so, it loads the default dissoc.dat file from ftpyfant/data/default
        directory and searches for the new abundances within self using the
        element symbol as key.

        The abundance values are "corrected" by subtracting 12.

        If element is not found in self, list.index() raises.
        """

        f = FileDissoc()
        f.init_default()
        for i, elem in enumerate(f.elems):
            if elem != " H":
                try:
                    j = self.ele.index(elem)
                    f.cclog[i] = self.abol[j]-12
                except ValueError:
                    # if dissoc element is not found in abonds, will use
                    # value in default dissoc.dat
                    pass
        return f

    def get_turbospectrum_str(self):
        """Returns ((atomic number, abundance), ...) string for TurboSpectrum.

        Returns a text string to be pasted in a script that runs TurboSpectrum.

        Elements whose symbol is not found in the periodic table are skipped
        without warning.
        """

        # determines the atomic numbers of the elements
        atomic_numbers, abunda = [], []
        for symbol, abundance in zip(self.ele, self.abol):
            s = symbol.strip()
            try:
                atomic_numbers.append("%3d" % (pyfant.SYMBOLS.index(s)+1))
                abunda.append(abundance)
            except ValueError:
                pass  # skips elements whose symbol is not in the periodic table
        # sorts by atomic number
        indexes = sorted(list(range(len(atomic_numbers))), key=lambda k: atomic_numbers[k])
        # mounts string
        l = ["'INDIVIDUAL ABUNDANCES:'   '%d'" % len(indexes), "  1 12.",]+\
            ["%s %g" % (atomic_numbers[i], abunda[i]) for i in indexes]
        return "\n".join(l)

    def sort_a(self):
        """Sorts alphabetically using self.ele.

        Symbols not found in the periodic table will appear first and  orderered
        alphabetically.
        """

        indexes = sorted(list(range(len(self))), key=lambda k: self.ele[k].strip())
        self.ele = [self.ele[i] for i in indexes]
        self.abol = [self.abol[i] for i in indexes]
        self.notes_per_ele = [self.notes_per_ele[i] for i in indexes]

    def sort_z(self):
        """
        Sorts by atomic number.

        Symbols not found in the periodic table will appear first.

        Returns: list with those symbols not found in the periodic table.
        """

        # first determines the atomic numbers of the elements
        atomic_numbers = []
        not_found = []
        for symbol in self.ele:
            s = symbol.strip()
            try:
                atomic_numbers.append("%3d" % (pyfant.SYMBOLS.index(s)+1))
            except ValueError:
                atomic_numbers.append("    "+s)
                not_found.append(symbol)

        indexes = sorted(list(range(len(atomic_numbers))), key=lambda k: atomic_numbers[k])
        self.ele = [self.ele[i] for i in indexes]
        self.abol = [self.abol[i] for i in indexes]
        self.notes_per_ele = [self.notes_per_ele[i] for i in indexes]

        return not_found


    # def make_dissoc(self):
    #     """Creates a new FileDissoc object.
    #
    #     Returns (file_dissoc, log), where
    #      - file_dissoc is a FileDissoc instance,
    #      - log is a list with this structure:
    #        [(element, dissoc abundance, message string), ...]
    #        if all atomic symbols are found in self, log will be an empty list.
    #        Otherwise it will contain the registry of the elements not found
    #        in self.
    #
    #     To do so, it loads the default dissoc.dat file from ftpyfant/data/default
    #     directory and searches for the new abundances within self using the
    #     element symbol as key.
    #
    #     The abundance values are "corrected" by subtracting 12.
    #     """
    #
    #     f, log = FileDissoc(), []
    #     f.init_default()
    #     for i, elem in enumerate(f.elems):
    #         try:
    #             j = self.ele.index(elem)
    #             f.cclog[i] = self.abol[j]-12
    #         except ValueError:
    #             log.append((elem, f.cclog[i],
    #              "Element \"%s\", kept original %g" % (elem, f.cclog[i])))
    #     return f, log





"""Python interface to the PFANT spectral synthesis software (Fortran)"""

# # Temporary imports
#   =================
# These modules should be be del'eted at the end
import a99


# # Setup
#   =====
SESSION_PREFIX_SINGULAR = 'session-'
SESSION_PREFIX_PLURAL = 'session-'
MULTISESSION_PREFIX = 'multi-session-'

def get_custom_multisession_dirname(session_id):
    """This defines how custom directory name is made up"""
    # return pyfant.MULTISESSION_PREFIX+session_id
    return session_id


# # Imports
#   =======
from .basic import *
from .filetypes import *
from .downloaders import *
from .run import *
from .util import *
from .convmol import *
from .from_vald import *
from .from_turbospectrum import *
from .molconsts import *
from .kovacs import *
from .gui import *
from . import gui
from .vis import *
from .misc import *


# # Function to access package-specific config file
#   ===============================================
def get_config():
    """Returns PyfantConfigObj object that corresponds to file ~/.ftpyfant.conf"""
    return a99.get_config_obj(".pyfant.conf")


# # Finally, gets rid of unwanted symbols in the workspace
#   ======================================================
del a99



"""
Constants shared among windows, and resources that are not general enough to be put in API
"""

# No need to implementa a "__all__", a99 will be del'ed at the end
import a99


DESCR_MULTI = """
Runs pfant for different abundances for each element, then run nulbad for each
pfant result for different FWHMs.

The configuration is read from a .py file.

The user must specify a list of FWHM values for nulbad convolutions, and
a dictionary containing element symbols and respective list containing n_abdif
differential abundances to be used for each element.

pfant will be run n_abdif times, each time adding to each element in ab the i-th
value in the vector for the corresponding element.

nulbad will run n_abdif*n_fwhms times, where n_fwhms is the number of different
FWHMs specified.

The result will be
- several spectra saved as  "<star name><pfant name or counter>.sp"
- several "spectra list" files saved as "cv_<FWHM>.spl". As the file indicates,
  each ".spl" file will have the names of the spectrum files for a specific FWHM.
  .spl files are subject to input for lineplot.py by E.Cantelli
"""


# Color for labels indicating a star parameter
COLOR_STAR = "#2A8000"
# Color for labels indicating a software configuration parameter
COLOR_CONFIG = "#BD6909"


# Messages shared in two or more different situations
INITIALIZES_SUN = "Initializes fields with default parameters (Sun)"
PARAMS_INVALID = "Can't save, invalid parameter values(s)!"
LLZERO_LLFIN = "The calculation interval for the synthetic spectrum is given by "\
  "["+a99.enc_name("llzero", COLOR_CONFIG)+", "+a99.enc_name("llfin", COLOR_CONFIG)+"]."\
"""
<pre>
          aint
          &lt;--&gt;

----[----|----|----|----|-]-------&gt;
    |                     |       wavelength (angstrom)
    llzero                llfin
</pre>
"""

DESCR_PTDISK = """
This option is used to simulate a spectrum acquired
out of the center of the star disk.<br><br>
This is useful if the synthetic spectrum will be compared with an
observed spectrum acquired out of the center of the star disk.
<ul>
    <li>True: 7-point integration
    <li>False: 6- or 26-point integration, depending on option --kik
</ul>"""


# Relating tablewidget column headers with set-of-lines attributes
# This is shared between XFileMolecules and XMolLinesEditor
SOL_HEADERS = ["lambda", "sj", "jj", "branch"]
SOL_HEADERS_PLOT = ["lambda", "sj", "jj"]
SOL_ATTR_NAMES = ["lmbdam", "sj", "jj", "branch"]


# Relating tablewidget column headers with Atom atributes
# This is shared between XFileAtoms and XAtomLinesEditor
ATOM_HEADERS = ["lambda", "kiex", "algf", "ch", "gr", "ge", "zinf"]
ATOM_ATTR_NAMES = ["lambda_", "kiex", "algf", "ch", "gr", "ge", "zinf"]


class PlotInfo(object):
    """Just a data container class, used by both the Atomic and Molecular lines editors"""
    def __init__(self):
        self.flag = True  # Whether the plot is supposed to be shown or not
        self.mpl_obj = None  # matplotlib Lines2D object
        self.axis = None  # matplotlib axis
        self.y_vector = None  # reference to sol.sj or jj

del a99


"""Window to edit both main and abundances"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from .a_WFileMain import *
from .a_WFileAbonds import *
from .a_WOptionsEditor import *
# from .a_XRunnableManager import XRunnableManager
import os
import matplotlib.pyplot as plt
import a99
import pyfant


__all__ = ["XMainAbonds"]


################################################################################
class XMainAbonds(a99.XLogMainWindow):
    """
    Args:
      parent=None: nevermind
      file_main (optional)-- FileMain instance
    """

    def __init__(self, parent=None, file_main=None, file_abonds=None):
        # State variables
        a99.XLogMainWindow.__init__(self, parent)
        # XRunnableManager instance
        self._manager_form = None
        # RunnableManager instance
        self._rm = None
        self.save_dir = None
        self.load_dir = None

        # # Synchronized sequences
        # Used in generic operations where only certain parameters change
        self.tab_texts = ["Main configuration (Alt+&1)",
                          "Abundances (Alt+&2)",
                          "Command-line options (Alt+&3)"]
        self.flags_changed = [False, False, False]
        self.save_as_texts = ["Save main configuration as...",
                              "Save abundances as...",
                              "Save command-line options as..."]
        self.open_texts = ["Load main configuration file",
                           "Load abundances file",
                           "Load command-line options file"]
        self.clss = [pyfant.FileMain, pyfant.FileAbonds, pyfant.FileOptions]
        self.wilds = ["*.dat", "*.dat", "*.py"]

        # # Menu bar
        b = self.menuBar()
        m = self.menu_file = b.addMenu("&File")

        ac = m.addAction("&Open...")
        ac.setShortcut("Ctrl+O")
        ac.triggered.connect(self.on_open)

        m.addSeparator()

        ac = m.addAction("&Save")
        ac.setShortcut("Ctrl+S")
        ac.triggered.connect(self.on_save)

        ac = m.addAction("Save &as...")
        ac.setShortcut("Ctrl+Shift+S")
        ac.triggered.connect(self.on_save_as)

        ac = m.addAction("Save a&ll...")
        ac.setShortcut("Alt+Shift+S")
        ac.triggered.connect(self.on_save_all)

        m.addSeparator()

        ac = m.addAction("Load &default")
        ac.setStatusTip("Loads default")
        ac.setShortcut("Ctrl+D")
        ac.triggered.connect(self.on_reset)

        m.addSeparator()

        ac = m.addAction("&Quit")
        ac.setShortcut("Ctrl+Q")
        ac.triggered.connect(self.close)

        m = self.menu_view = b.addMenu("&View")

        ac = m.addAction("Runnables &Manager")
        ac.setShortcut("Ctrl+M")
        ac.triggered.connect(self.on_show_rm)


        # # It is necessary to do a little more work to create this option
        # ac = m.addAction("&Explorer")
        # ac.setShortcut("Ctrl+E")
        # ac.triggered.connect(self.on_show_explorer)

        # # Central layout

        cw = self.centralWidget = QWidget()
        self.setCentralWidget(cw)
        la = self.layout_main = QVBoxLayout(cw)

        # ## Main control bar
        # A layout is created and left blank for descendants to add
        # widgets as needed

        l = self.controlLayout = QHBoxLayout()
        la.addLayout(l)

        # ## Tabs

        # tab "File"
        tt = self.tabWidget = QTabWidget(self)
        la.addWidget(tt)
        tt.setFont(a99.MONO_FONT)

        # ### Main configuration tab
        w0 = self.c27272 = QWidget()
        tt.addTab(w0, self.tab_texts[0])
        l0 = self.c12842 = QVBoxLayout(w0)

        # #### Main configuration file toolbar
        l1 = self.c29378 = QHBoxLayout()
        l0.addLayout(l1)
        w = QLabel("<b>File:<b>")
        l1.addWidget(w)
        w = self.label_fn_main = QLabel()
        l1.addWidget(w)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # #### Main file editor widget
        me = self.me = WFileMain()
        l0.addWidget(me)
        me.changed.connect(self.on_main_edited)

        # ### Abundances tab
        w0 = self.c10101 = QWidget()
        tt.addTab(w0, self.tab_texts[1])
        l0 = self.c54354 = QVBoxLayout(w0)

        # #### Abundances file toolbar
        l1 = self.c65478 = QHBoxLayout()
        w = QLabel("<b>File:</b>")
        l0.addLayout(l1)
        l1.addWidget(w)
        w = self.label_fn_abonds = QLabel()
        l1.addWidget(w)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # #### Abundances editor
        ae = self.ae = WFileAbonds()
        l0.addWidget(ae)

        ae.changed.connect(self.on_abonds_edited)

        # ### Command-line options tab
        w0 = QWidget()
        tt.addTab(w0, self.tab_texts[2])
        l0 = QVBoxLayout(w0)

        # # #### File label
        # l1 = self.c293wd = QHBoxLayout()
        # l0.addLayout(l1)
        # w = QLabel("<b>File:<b>")
        # l1.addWidget(w)
        # w = self.label_fn_options = QLabel()
        # l1.addWidget(w)
        # l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        oe = self.oe = WOptionsEditor(self)
        oe.changed.connect(self.on_options_edited)
        l0.addWidget(oe)
        #        tt.tabBar()

        # ### Final tabs setup
        tt.setCurrentIndex(0)

        # ### These sequences couldn't be set above because the widgets didn't exist yet
        self.editors = [self.me, self.ae, self.oe]
        self.labels_fn = [self.label_fn_main, self.label_fn_abonds, self.oe.label_fn]
        assert len(self.tab_texts) == len(self.flags_changed) == \
               len(self.save_as_texts) == len(self.open_texts) == len(self.clss) == \
               len(self.editors) == len(self.labels_fn)

        # # Loads default files
        if os.path.isfile(pyfant.FileMain.default_filename):
            f = pyfant.FileMain()
            f.load()
            self.me.load(f)
        if os.path.isfile(pyfant.FileAbonds.default_filename):
            f = pyfant.FileAbonds()
            f.load()
            self.ae.load(f)
        if os.path.isfile(pyfant.FileOptions.default_filename):
            f = pyfant.FileOptions()
            f.load()
            self.oe.load(f)
        else:
            self.oe.load(pyfant.FileOptions())
        self._update_labels_fn()



    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Qt override

    def closeEvent(self, event):
        flag_exit, ff = True, []
        for ed, flag_changed in zip(self.editors, self.flags_changed):
            if ed and ed.f and flag_changed:
                ff.append(ed.f.description)

        if len(ff) > 0:
            s = "Unsaved changes\n  -"+("\n  -".join(ff))+"\n\nAre you sure you want to exit?"
            flag_exit = a99.are_you_sure(True, event, self, "Unsaved changes", s)
        if flag_exit:
            plt.close("all")

    def keyPressEvent(self, evt):
        incr = 0
        if evt.modifiers() == Qt.ControlModifier:
            n = self.tabWidget.count()
            if evt.key() in [Qt.Key_PageUp, Qt.Key_Backtab]:
                incr = -1
            elif evt.key() in [Qt.Key_PageDown, Qt.Key_Tab]:
                incr = 1
            if incr != 0:
                new_index = self.__get_index()+incr
                if new_index < 0:
                    new_index = n-1
                elif new_index >= n:
                    new_index = 0
                self.tabWidget.setCurrentIndex(new_index)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for Qt library signals

    # Index checks are because descendant classes may add more tabs

    def on_open(self):
        if self.__tab_has_file_operations():
            self.__generic_open()

    def on_save(self):
        if self.__tab_has_file_operations():
            self.__generic_save()

    def on_save_as(self):
        if self.__tab_has_file_operations():
            self.__generic_save_as()

    def on_save_all(self):
        index_save = self.__get_index()
        try:
            for index in range(self.tabWidget.count()):
                self.tabWidget.setCurrentIndex(index)
                if not self.__generic_save():
                    # breaks if user cancels a "save as" operation
                    break
        finally:
            self.tabWidget.setCurrentIndex(index_save)

    def on_reset(self):
        if self.__tab_has_file_operations():
            idx = self.__get_index()
            editor = self.editors[idx]
            flag_ok = True
            if editor.f:
                descr = self.__get_tab_description()
                r = QMessageBox.question(self, "Load default", "Current setup "
                 "for %s will be overwritten with a 'default' setup.\n\n"
                 "Confirm?" % descr, QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
                flag_ok = r == QMessageBox.Yes
            if flag_ok:
                self.__generic_reset()

    def on_show_rm(self):
        if self._manager_form:
            self._manager_form.show()
            self._manager_form.raise_()
            self._manager_form.activateWindow()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for signals emited by ftpyfant widgets

    def on_main_edited(self):
        self._on_changed()

    def on_abonds_edited(self):
        self._on_changed()

    def on_options_edited(self):
        self._on_changed()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Protected methods to be overriden or used by descendant classes

    def _on_changed(self):
        index = self.__get_index()
        self.flags_changed[index] = True
        self.__update_tab_texts()

    def _check_single_setup(self):
        """Checks if setup parameters are valid.

        The verifications here apply both to single and multi mode.
        """
        errors = []
        if not self.me.f:
            errors.append("main configuration not set")
        else:
            if not self.me.flag_valid:
                errors.append("error(s) in main configuration")

        if not self.ae.f:
            errors.append("abundances not set")
        else:
            if not self.ae.flag_valid:
                errors.append("error(s) in abundances")
        if not self.oe.flag_valid:
            errors.append("error(s) in command-line options")
        return errors

    def _update_labels_fn(self):
        for editor, label in zip(self.editors, self.labels_fn):
            if not label:
                continue
            if not editor.f:
                text = "(not loaded)"
            elif editor.f.filename:
                text = os.path.relpath(editor.f.filename, ".")
            else:
                text = "(filename not set)"
            label.setText(text)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Gear

    def __get_index(self):
        """Returns index of current selected tab."""
        return self.tabWidget.currentIndex()

    def __get_tab_description(self):
        """Returns "description" of current tab (tab text without shortcut info)."""
        idx = self.__get_index()
        text = self.tab_texts[idx]
        if "(" in text:
            text = text[:text.index("(")-1]
        text = text[0].lower() + text[1:]
        return text

    def __generic_reset(self):
        index = self.__get_index()
        editor, text, cls = self.editors[index], self.open_texts[index], self.clss[index]
        f = cls()
        f.init_default()
        editor.load(f)
        self.flags_changed[index] = True
        self.__update_tab_texts()
        self._update_labels_fn()

    def __generic_save(self):
        """Returns False if user has cancelled a "save as" operation, otherwise True."""
        index = self.__get_index()
        editor = self.editors[index]
        f = editor.f
        if not f:
            return True
        if not editor.flag_valid:
            a99.show_error("Cannot save, %s has error(s)!" % f.description)
        if f.filename:
            try:
                f.save_as()
                self.flags_changed[index] = False
                self.__update_tab_texts()
                return True
            except Exception as e:
                a99.show_error(str(e))
                raise
        else:
            return self.__generic_save_as()

    def __generic_save_as(self):
        """Returns False if user has cancelled operation, otherwise True."""
        index = self.__get_index()
        editor, text, wild = self.editors[index], self.save_as_texts[index], self.wilds[index]
        if not editor.f:
            return True
        if editor.f.filename:
            d = editor.f.filename
        else:
            d = os.path.join(self.save_dir if self.save_dir is not None
                             else self.load_dir if self.load_dir is not None
                             else ".", editor.f.default_filename)
        new_filename = QFileDialog.getSaveFileName(self, text, d, wild)[0]
        if new_filename:
            self.save_dir, _ = os.path.split(str(new_filename))
            try:
                editor.f.save_as(str(new_filename))
                self.flags_changed[index] = False
                self._update_labels_fn()
                self.__update_tab_texts()
                return True
            except Exception as e:
                a99.show_error(str(e))
                raise
        return False

    def __generic_open(self):
        index = self.__get_index()
        editor = self.editors[index]
        text = self.open_texts[index]
        cls = self.clss[index]
        label = self.labels_fn[index]

        # makes wildcard options a list and adds "*" if not present
        SEP = ";;"
        __wild = self.wilds[index]
        _wild = [x.strip() for x in __wild.split(SEP)]
        if "*" not in _wild or "*.*" not in _wild:
            _wild.append("*")
        wild = SEP.join(_wild)

        try:
            d = self.load_dir if self.load_dir is not None \
                else self.save_dir if self.save_dir is not None \
                else "."
            new_filename = QFileDialog.getOpenFileName(self, text, d, wild)[0]
            if new_filename:
                self.load_dir, _ = os.path.split(str(new_filename))
                f = cls()
                f.load(str(new_filename))
                editor.load(f)
                self._update_labels_fn()
                self.__update_tab_texts()
        except Exception as e:
            a99.show_error(str(e))
            raise

    def __tab_has_file_operations(self):
        return self.flags_changed[self.__get_index()] is not None

    def __update_tab_texts(self):
        for index, (text, flag_changed) in enumerate(zip(self.tab_texts, self.flags_changed)):
            self.tabWidget.setTabText(index,
             text+(" (changed)" if flag_changed else ""))



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
import f311
import pyfant

__all__ = ["XFileMolDB"]


class XFileMolDB(f311.XFileMainWindow):
    def _add_stuff(self):

        me = self.w_moldb = pyfant.WFileMolDB(self)
        me.changed.connect(self._on_changed)

        # # Synchronized sequences
        self.pages.append(f311.MyPage(
         text_tab=pyfant.FileMolDB.description,
         cls_save=pyfant.FileMolDB, clss_load=(pyfant.FileMolDB,), wild="*.sqlite",
         editor=me, flag_autosave=True
        ))

        self.setWindowTitle("Molecular information database editor")

    def wants_auto(self):
        idx = self.w_source.index
        filename = None
        if idx == 0:
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = a99.new_filename("mol", "dat")

        self.w_out.value = filename



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from a99 import WDBRegistry
import a99
import pyfant
from .a_WDBSystem import *

__all__ = ["WDBFCF"]


# Field names to leave out of table widget
_FIELDNAMES_OUT = ("id", "id_system",)

# TRAPRB input parameters to be edited before TRAPRB is run
_TRAPRB_ATTRS = ["rmin", "rmax", "delr", "maxv"]

class WDBFCF(WDBRegistry):
    """Registry for table 'fcf'"""

    def __init__(self, *args):
        WDBRegistry.__init__(self, *args)

        self._id_system = None

        # Adds button to download NIST data
        action = self.action_download = QAction(a99.get_icon("alien"),
            "Run &TRAPRB to get Franck-Condon Factors...\n\n"
            "For more information on TRAPRB input variables, visit URL:\n\n"
            "https://github.com/trevisanj/pyfant/blob/master/docs/art/traprb-input.pdf", self)
        # action.setToolTip("")
        action.triggered.connect(self._on_run_traprb)
        self.toolbar.addAction(action)

    def set_id_system(self, id_):
        """Sets molecule id and re-populates table"""
        self._id_system = id_
        self._populate()
        self._move_to_first()

    def _on_run_traprb(self):
        if self._id_system is None:
            return

        row = self._f.get_conn().execute("select * from system where id = ?",
                                         (self._id_system,)).fetchone()

        string = pyfant.molconsts_to_system_str(row, SYSTEMSTYLE)



        n = len(self._data)
        beware = "" if n == 0 else \
            "\n\n**Attention**: {} existing FCF{} for this system will be deleted!".\
                format(n, "" if n == 1 else "s")

        msg = "TRAPRB will be executed to calculate the Franck-Condon factors for system '{}'." \
              "{}\n\nDo you wish to continue?".format(string, beware)

        r = QMessageBox.question(self, "Calculate FCFs", msg,
                                 QMessageBox.Yes | QMessageBox.No,
                                 QMessageBox.Yes if n == 0 else QMessageBox.No)

        if r == QMessageBox.Yes:
            r, form = self._show_traprb_parameters_form(r)

            if r == QDialog.Accepted:
                kwargs = form.get_kwargs()

                mc = pyfant.MolConsts()

                try:
                    mc.populate_all_using_system_row(self._f, row)

                    # TODO maybe intermediate review stage, i.e., allow the user to see the FCFs before inserting

                    traprb = pyfant.run_traprb(mc, None, None, **kwargs)

                    traprb.load_result()
                    output = traprb.result["output"]

                    pyfant.insert_fcfs(self._f, self._id_system, output.fcfs, flag_replace=True)

                    self._populate()

                    n = len(output.fcfs)
                    self.add_log("{} FCF{} successfully inserted".
                                 format(n, "" if n == 1 else "s"), True)
                except Exception as e:
                    self.add_log_error("Error calculating FCFs: '{}'".format(a99.str_exc(e)),
                                       True, e)

    def _show_traprb_parameters_form(self, r):
        # **Note** creates TRAPRBInputState() instance to get default input parameter values

        r, form = a99.show_edit_form(
            pyfant.TRAPRBInputState(),
            attrs=_TRAPRB_ATTRS,
            title="Modify selected TRAPRB input parameters",
            toolTips=[
                "minimum radius for calculation of wavefunctions (0.75 in most cases; 0.6 for OH A-X)",
                "maximum radius for calculation of wavefunctions",
                "radial interval for calculation of wavefunctions",
                "maximum quantum number for calculation of results"],
        )
        return r, form

    def _populate(self, restore_mode=None):
        """
        Populates table widget

        Args:
            restore_mode: how to reposition cursor:
                - "index": tries to restore same row index
                - anything else: does not re-position
        """

        if self._f is None:
            return

        self._flag_populating = True
        try:
            curr_idx = self.tableWidget.currentRow()

            t = self.tableWidget
            rows = a99.cursor_to_rows(self._f.query_fcf(id_system=self._id_system))
            ti = self._f.get_table_info("fcf")
            fieldnames = [name for name in ti if name not in _FIELDNAMES_OUT]
            # unfortunately QTableWidget is not prepared to show HTML col_names = [row["caption"] or row["name"] for row in ti if not row["name"] in FIELDNAMES_OUT]
            col_names = fieldnames
            nr, nc = len(rows), len(fieldnames)

            a99.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(col_names)

            if nr > 0:
                for i, row in enumerate(rows):
                    for j, fieldname in enumerate(fieldnames):
                        cell = row[fieldname]
                        item = QTableWidgetItem("" if cell is None else str(cell))
                        t.setItem(i, j, item)
            t.resizeColumnsToContents()
            self._data = rows

            if restore_mode == "index":
                if -1 < curr_idx < t.rowCount():
                    t.setCurrentCell(curr_idx, 0)
        finally:
            self._flag_populating = False
            self._wanna_emit_id_changed()

    def _get_edit_params(self):
        """Returns a Parameters object containing information about the fields that may be edited"""
        ti = self._f.get_table_info("fcf")
        params = a99.table_info_to_parameters(ti)
        params = [p for p in params if not p.name.startswith("id")]
        return params

    def _get_edit_field_names(self):
        """Returns a list with the names of the fields that may be edited"""
        params = self._get_edit_params()
        ret = [p.name for p in params]
        return ret

    def _do_on_insert(self):
        if self._id_system is None:
            return
        params = self._get_edit_params()
        form = a99.XParametersEditor(specs=params, title="Insert Franck-Condon Factor")
        r = form.exec_()
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            conn = self._f.get_conn()
            s = "insert into fcf (id_system, {}) values ({}, {})".\
                format(", ".join([p.name for p in params]),
                       self._id_system,
                       ", ".join(["?"]*len(params)))
            cursor = conn.execute(s, list(kwargs.values()))
            id_ = cursor.lastrowid
            conn.commit()
            self._populate(False)
            self._find_id(id_)
            return True

    def _do_on_edit(self):
        r, form = a99.show_edit_form(self.row, self._get_edit_field_names(), "Edit Franck-Condon Factor")
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            id_ = self.row["id"]
            s = "update fcf set {} where id = {}".format(
                ", ".join(["{} = '{}'".format(a, b) for a, b in kwargs.items()]), id_)
            conn = self._f.get_conn()
            conn.execute(s)
            conn.commit()
            self._populate()
            self._find_id(id_)
            return True

    def _do_on_delete(self):
        r = QMessageBox.question(None, "Delete Franck-Condon Factor",
                                 "Are you sure?",
                                 QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if r == QMessageBox.Yes:
            conn = self._f.get_conn()
            id_ = self.row["id"]
            conn.execute("delete from fcf where id = ?", [id_])
            conn.commit()

            self._populate("index")
            return True



#!/usr/bin/env python

"""
Looks for files "*.norm" inside directories session-* and saves one figure per page in a PDF file

References:
http://stackoverflow.com/questions/17788685
"""

import matplotlib.backends.backend_pdf
import matplotlib.pyplot as plt
import a99
import glob
import os
import argparse
import pyfant
import logging
import f311


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )

    parser.add_argument('--samey',
        help='Creates all plots with same y-limits of [0, 1.02]', action="store_true")
    parser.add_argument('fn_output', nargs='?', default='output.pdf', type=str,
                        help='PDF output file name')

    args = parser.parse_args()

    dd = glob.glob(pyfant.SESSION_PREFIX_SINGULAR+"*")
    dd.sort()

    pdf = matplotlib.backends.backend_pdf.PdfPages(args.fn_output)

    a99.format_BLB()

    for d in dd:
        name = d[8:]

        a99.get_python_logger().info("Looking into session '{}'...".format(name))

        norm_filenames = glob.glob(os.path.join(d, "*.norm"))
        if len(norm_filenames) == 0:
            continue

        for filename in norm_filenames:
            a99.get_python_logger().info("    File '{}'".format(filename))
            f = pyfant.FileSpectrumPfant()

            # Note: takes first .norm file that finds
            f.load(filename)

            f311.draw_spectra_stacked([f.spectrum])  #v.title = "%s" % name

            fig = plt.gcf()
            if args.samey:
                plt.ylim([0, 1.02])
            else:
                y = f.spectrum.y
                ymin, ymax = min(y), max(y)
                margin = .02*(ymax-ymin)
                plt.ylim([ymin-margin, ymax+margin])
            ax = plt.gca()
            p = ax.get_position()
            p.x0 = 0.11
            ax.set_position(p)  # Try to apply same position for all figures to improve flicking experience
            pdf.savefig(fig)
            plt.close()

    # for fig in xrange(1, figure().number): ## will open an empty extra figure :(
    #     pdf.savefig( fig )
    pdf.close()




import pyfant


def test_greek_to_spdf():
    tt = [pyfant.greek_to_spdf(x) for x in pyfant.basic.conversion._SPDF_TO_GREEK.values()]
    assert(tt == list(pyfant.basic.conversion._GREEK_TO_SPDF.values()))


def test_spdf_to_greek():
    tt = [pyfant.spdf_to_greek(x) for x in range(len(pyfant.basic.conversion._SPDF_TO_GREEK))]
    assert(tt == list(pyfant.basic.conversion._SPDF_TO_GREEK.values()))


def test_parse_system_str():
    assert pyfant.parse_system_str("[d 3 phi - a 2 delta]") == ["d", 3, 3, "a", 2, 2]



"""
Conversion of molecular lines files.
"""

from .convlog import *
from .calc_qgbd import *
from .conv_vald3 import *
from .conv_kurucz import *
from .conv import *
from .conv_plez import *
# from .conv_hitran import *
from .conv_hitran160 import *
from .conv_brooke2014 import *


"""
Searches molecular lines around given wavelength interval

Included inverval will be central_lambda-tolerance <= line_wavelength <= central_lambra+tolerance.
"""

import argparse, logging, a99, pyfant, tabulate


a99.logging_level = logging.INFO
a99.flag_log_file = False


# Searching which molecule and (v', v'') has line with given wavelength
import pyfant, f311, a99, a107

# FILENAME = "molecules.dat"
# FILENAME = "CO_dV11_stable.PFANT.dat"
# FILENAME = "CN1214-Brookeetal-2014-list.PFANT.dat"
# LAMBDA = 16388.6
# FILENAME = "molecule-CNredAX.dat"
FILENAME = "molecules.dat"
# LAMBDA = 15577.42
# LAMBDA = 15568.8
LAMBDA = 15328.5

TOLERANCE = 1

def pprint(s):
    for line in s.split("\n"):
        print(line)


def main(args):
    central_lambda = args.central_lambda
    tolerance = args.tolerance
    l0 = central_lambda-tolerance
    l1 = central_lambda+tolerance

    f = pyfant.FileMolecules()
    f.load(args.fn_input)

    mols = []
    rows = []
    header = ["Molecule #", "Set-of-lines #", "vl", "v2l", "lmbdam", "sj", "jj", "branch"]
    floatfmts = ["", "", "", "", ".4f", "", "", ""]

    for i_m, mol in enumerate(f.molecules):
        found = False
        for i_s, sol_ in enumerate(mol.sol):
            for i, (lmbdam, sj, jj, branch) in enumerate(zip(sol_.lmbdam, sol_.sj, sol_.jj, sol_.branch)):
                if l0 <= lmbdam <= l1:
                    rows.append([i_m+1, i_s+1, sol_.vl, sol_.v2l, lmbdam, sj, jj, branch])

                    if not found:
                        mols.append([i_m+1, mol.description])
                        found = True

    pprint("\n"+"\n".join(a99.format_box("Lines found")))
    pprint(tabulate.tabulate(rows, header, floatfmt=floatfmts))
    pprint("\n"+"\n".join(a99.format_box("Molecules indexes from table above")))
    pprint(tabulate.tabulate(mols, ["Molecule #", "Description"]))




if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('central_lambda', type=float,
     help='Central wavelength (angstrom)')
    parser.add_argument('fn_input', type=str, help='input file name')
    parser.add_argument('-t', "--tolerance", type=float, default=0.5,
                        help='Wavelength tolerance around central wavelength (left and right)')

    args = parser.parse_args()

    main(args)


__all__ = ["XAtomLinesEditor"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from ._shared import *
import a99
import pyfant


class XAtomLinesEditor(QMainWindow):

    def __init__(self, parent):
        QMainWindow.__init__(self)

        # objects in matplotlib window

        self.parent = parent
        self.flag_populating = False  # activated when populating table

        a = self.tableWidget = QTableWidget()
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)
        a.cellChanged.connect(self.on_tableWidget_cellChanged)
        a.setEditTriggers(QAbstractItemView.DoubleClicked | QAbstractItemView.EditKeyPressed)
        a.setFont(a99.MONO_FONT)
        a.installEventFilter(self)


        self.setCentralWidget(a)
        a99.snap_right(self, 360)

    def on_tableWidget_currentCellChanged(self, currentRow, currentColumn, previousRow,
                                          previousColumn):
        self.parent.AtomLinesEditor_current_row_changed(currentRow)

    def on_tableWidget_cellChanged(self, row, column):
        if not self.flag_populating:
            item = self.tableWidget.item(row, column)
            try:
                value = float(item.text())
            except ValueError:
                # restores original value
                a99.show_error("Invalid floating point value: %s" % item.text())
                item.setText(str(self.parent.atom.__getattribute__(ATOM_ATTR_NAMES[column])[row]))
            else:
                self.parent.AtomLinesEditor_cell_changed(row, column, value)

    def closeEvent(self, _):
        self.parent.AtomLinesEditor_closing()

    def eventFilter(self, source, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Return:
                if source == self.tableWidget:
                    self.tableWidget.editItem(self.tableWidget.currentItem())
                    return True
        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # *

    def set_atom(self, atom, title):
        """Sets set of lines."""

        assert isinstance(atom, pyfant.Atom)

        self.flag_populating = True
        try:
            self.setWindowTitle(title)

            t = self.tableWidget
            n = len(atom)
            a99.reset_table_widget(t, n, len(ATOM_HEADERS))
            t.setHorizontalHeaderLabels(ATOM_HEADERS)

            # list with the vectors themselves
            attrs = [atom.__getattribute__(x) for x in ATOM_ATTR_NAMES]

            for i in range(len(atom)):
                for j, attr in enumerate(attrs):
                    item = QTableWidgetItem(str(attr[i]))
                    t.setItem(i, j, item)

            t.resizeColumnsToContents()
        finally:
            self.flag_populating = False


    def set_row(self, i):
        t = self.tableWidget
        c = t.currentColumn()
        self.tableWidget.setCurrentCell(i, c)



"""VALD3 atomic or molecular lines file"""


__all__ = ["FileVald3", "Vald3MolecularLine", "Vald3SolKey"]

# from ..gear import *
import sys
import a99
from f311 import DataFile
import io
from collections import namedtuple, defaultdict

#: List of all atomic symbols
_symbols = [
'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si',
 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co',
 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr',
 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I',
 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au',
 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U',
 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db',
 'Sg', 'Bh', 'Hs', 'Mt'
]


@a99.froze_it
class Vald3Species(a99.AttrsPart):
    """
    Represents species of VALD3 **extended format** atomic/molecular lines file

    Examples: "OH 1", "Fe 2"

    """
    attrs = ["formula", "ioni"]

    # # Properties that iterate through the Vald3Line objects to mount vectors
    # @property
    # def lambda_(self):
    #     return np.array([x.lambda_ for x in self.lines])

    def __init__(self):
        a99.AttrsPart.__init__(self)
        self.lines = []  # list of Vald3Line
        self.formula = None
        self.ioni = None

    def __len__(self):
        return len(self.lines)

    def __str__(self):
        return "{} {}".format(self.formula, self.ioni)

    def __repr__(self):
        return "'%s%s'" % (self.formula, self.ioni)

    def __iter__(self):
        return iter(self.lines)

    def _cut(self, llzero, llfin):
        """Keeps only the lines with their llzero <= lambda_ <= llfin."""
        for i in reversed(list(range(len(self)))):
            if not (llzero <= self.lines[i].lambda_ <= llfin):
                del self.lines[i]


# TODO Today I leave it like this, but tomorrow I must group the (system+vl,v2l)

# @a99.froze_it
# class Vald3Line(a99.AttrsPart):
#     attrs = ["lambda_", "loggf", "Jl", "J2l", "vl", "v2l"]
#
#     def __init__(self):
#         a99.AttrsPart.__init__(self)
#
#         self.lambda_ = None
#         self.loggf = None
#         self.Jl = None
#         self.J2l = None
#         self.vl = None
#         self.v2l = None
#         self.sys0 = None
#         self.sys1 = None
#         self.sys2 = None


Vald3MolecularLine = namedtuple("Vald3MolecularLine", ["lambda_", "loggf", "Jl", "J2l"])
Vald3SolKey = namedtuple("Vald3SolKey", ["vl", "v2l", "sys00", "sys01", "sys02", "sys10", "sys11", "sys12"])

class FileVald3(DataFile):
    """
    VALD3 atomic or molecular lines file

    Args:
        flag_parse_atoms=True: whether or not to parse the atomic lines
        flag_parse_molecules=True: whether or not to parse the molecular lines


    Species are encoded as follows.

    self.speciess (double s) is a dictionary where

        key: (formula, ionization), e.g. ("MgH", 1)
        value: "sets of lines", dict where

            key: 0 (int) if atom (not implemented)
                 (vl, v2l, sys00, sys01, sys02, sys10, sys11, sys12) if molecule
            value: list of Vald3AtomicLine if atom (not implemented yet)
                   list of Vald3MolecularLine if molecule

    TODO: parsing of atomic lines!

    **Note** Load only

    **Note** Does not load all file fields, only those needed to convert this file to
             PFANT molecular lines format (20161216)
    """

    attrs = ["speciess", "num_lines"]

    @property
    def num_lines(self):
        ret = sum(map(len, self.speciess))
        return ret

    def __len__(self):
        return len(self.speciess)

    def __init__(self, flag_parse_atoms=True, flag_parse_molecules=True):
        DataFile.__init__(self)

        # Configuration
        self.flag_parse_atoms = flag_parse_atoms
        self.flag_parse_molecules = flag_parse_molecules

        self.speciess = {}

    def __iter__(self):
        return iter(self.speciess)

    def _do_load(self, filename):
        """Clears internal lists and loads from file."""

        with open(filename, "r") as h:
            self._do_load_h(h, filename)

    def _do_load_h(self, h, filename):
        # ---SAMPLE FILE---
        #                                                                    Lande factors      Damping parameters
        # Elm Ion      WL_air(A)   log gf* E_low(eV) J lo  E_up(eV) J up  lower  upper   mean   Rad.  Stark  Waals
        # 'OH 1',       16401.409,  -8.587,  4.0602, 39.5,  4.8160, 39.5,99.000,99.000,99.000, 0.000, 0.000, 0.000,
        # '  Hb                                                     2s2.3s2.1p3          X,2,1,f,40,3'
        # '  Hb                                                 2s2.3s2.1p3              X,2,1,f,39,7'

        r = 0  # counts rows of file
        self.speciess = defaultdict(lambda: defaultdict(lambda: []))
        try:
            # Uses header as "magic characters"
            s = h.readline().strip("\n")
            if s != "                                                                   Lande factors      Damping parameters":
                raise RuntimeError("This doesn't appear to be a VALD3 extended-format atomic/molecular lines file")
            r += 1
            h.readline()
            r += 1

            while True:
                # Line 1/4
                # ========
                #
                s = h.readline().strip()
                if len(s) == 0 or not s.startswith("'"):
                    # EOF: blank line or references section
                    break

                fields = s.split(",")
                formula, s_ioni = fields[0][1:-1].split(" ")

                if formula in _symbols:
                    r += 1
                    # Atom, skips altogether (skips next three lines and loops)
                    for _ in range(3):
                        h.readline()
                        r += 1
                    continue

                line = Vald3MolecularLine(lambda_=float(fields[1]),
                                          loggf=float(fields[2]),
                                          Jl=float(fields[6]),  # J_up
                                          J2l = float(fields[4])  # J_lo
                                          )
                r += 1

                # Line 2/4 and 3/4
                # ================
                #
                # Parses "molecular system" TODO rename this term later
                # TODO dunno what comes first, vl or v2l, but if it follows the file header,
                # v2l (v lower) should come first, because "J lo" comes before "J up" in the file header
                s = h.readline().strip()
                r += 1
                sys00, sys01, sys02, v2l = _parse_system(s)
                s = h.readline().strip()
                r += 1
                sys10, sys11, sys12, vl = _parse_system(s)

                species_key = (formula, int(s_ioni))
                sol_key = Vald3SolKey(vl, v2l, sys00, sys01, sys02, sys10, sys11, sys12)
                self.speciess[species_key][sol_key].append(line)

                # Line 4/4 (citation line (skips))
                # ================================
                #
                h.readline()
                r += 1

        except Exception as e:
            raise type(e)(("Error around %d%s row of file '%s'" %
                           (r + 1, a99.ordinal_suffix(r + 1), filename)) + ": " + str(
                e)).with_traceback(sys.exc_info()[2])

def _parse_system(s):
    """Parses "system" information from VALD3 string

    Args:
        s: string (see examples below)

    Examples of lines:

    ```
    '  Hb                                                  3s2.4s2.1p4.5s1          X,2,0,,19,3'
    '  Hb                                                 3s2.4s2.1p3.5s2          A,2,1,e,20,9'
    '  Hb                         2sg2.2su2.3sg1.1pu3                            a,3,1,e,29,3,g'
    '  Hb                         2sg2.2su1.3sg1.1pu3/2sg2.2su2.3sg1.1pg1.1pu2   d,3,1,e,29,2,g'
    ```

    Returns:
        tuple: (sys0, sys1, sys2, v), wheter

            - sys0, sys1, sys2: information about system, TODO rename when I know better their meanings);
            - v: quantum number, either vl or v2l
    """

    kk = s[s.find(",")-1:-1].split(",")
    ret = {}
    sys0 = kk[0]
    sys1 = int(kk[1])
    sys2 = int(kk[2])
    try:
        v = int(kk[-1])
    except ValueError:
        # This covers the case when last item is "g" (see example in docstring)
        v = int(kk[-2])
    return sys0, sys1, sys2, v


def _fake_file():
    """Returns StringIO file to test FileVald3 class"""

    f = io.StringIO()
    f.write("""                                                                   Lande factors      Damping parameters
Elm Ion      WL_air(A)   log gf* E_low(eV) J lo  E_up(eV) J up  lower  upper   mean   Rad.  Stark  Waals
'Fe 1',       16400.045,  -4.131,  6.5922,  6.0,  7.3480,  5.0, 1.320, 1.090, 1.910, 8.140,-3.840,-7.330,
'  LS                                                                       3d7.(4F).4d f5G'
'  JK                                                             3d7.(4F<7/2>).4f 2[11/2]*'
'K14               Kurucz Fe I 2014   1 K14       1 K14       1 K14       1 K14       1 K14       1 K14       1 K14       1 K14       1 K14     Fe            '
'CO 1',       16400.058,  -5.951,  1.3820, 38.0,  2.1378, 39.0,99.000,99.000,99.000, 2.000, 0.000, 0.000,
'  Hb                                                4s2.5s2.1p4              X,2,0,,none,4'
'  Hb                                                    4s2.5s2.1p4          X,2,0,,none,7'
'KCO                                  2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO     (12)C(16)O    '
'CO 1',       16400.090,  -8.995,  1.1482, 53.0,  1.9040, 54.0,99.000,99.000,99.000, 2.000, 0.000, 0.000,
'  Hb                                                4s2.5s2.1p4              X,2,0,,none,2'
'  Hb                                                    4s2.5s2.1p4          X,2,0,,none,5'
'KCO                                  2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO     (12)C(18)O    '
""")

    f.seek(0)
    return f

def _test():
    """Test code in docstring

    Example:

    >>> h = _fake_file()
    >>> f = FileVald3()
    >>> f._do_load_h(h, "_fake_file")
    """
    return

# h = _fake_file()
# f = FileVald3()
# f._do_load_h(h, "_fake_file")






#!/usr/bin/env python

"""
Splits PFANT molecular linelist into several files, 1 file for each molecule

Output files will be named as "<original base name>-split-<molecule description>[-<n>].dat" (<n> is a number only present if
there are two or more molecules with the same description.

**Note**: existing files with same name as any output files are overwritten. This should not be a problem,
          since output filenames derive from the input filename, so there is little risk of overwritting something
          unrelated. If necessary, the input file could be isolated in a separate directory.
"""

import pyfant
import argparse
import os
import glob
import sys
import logging
import a99
import re


a99.logging_level = logging.INFO
a99.flag_log_file = False

def pprint(s):
    a99.get_python_logger().info(s)


def slugify(s):
    """Converts molecular description into something that can be used as filename.

    Steps:
        - calls a99.slugify()
        - Replaces multiple dashes with single dash
    """
    ret = a99.slugify(s)
    ret = re.sub("(-+)", "-", ret)
    return ret

def main(fn_input):
    f = pyfant.FileMolecules()
    f.load(fn_input)

    prefix = os.path.splitext(fn_input)[0]

    basenames = []
    for mol in f.molecules:
        fo = pyfant.FileMolecules()
        fo.molecules = [mol]
        fo.titm = f"Generated by split-molecules.py from '{fn_input}'"

        # Makes filename, caring for repeating descriptions
        slug = slugify(mol.description)
        _basename = basename = f"{prefix}-split-{slug}"
        n = 0
        while basename in basenames:
            n += 1
            if n > 99:
                raise RuntimeError("Too many molecules having the same description")
            basename = f"{_basename}-{n:02}"
        basenames.append(basename)

        fn_output = f"{basename}.dat"

        if os.path.isfile(fn_output):
            a99.get_python_logger().warning(f"Overwritting file '{fn_output}'")

        fo.save_as(fn_output)
        pprint(f"Saved file '{fn_output}'")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument("fn_input", type=str, help="Input filename")

    args = parser.parse_args()

    main(args.fn_input)



"""
Class to store all command-line options & DataFile instances used to run one or more
executables.
"""

__all__ = ["Conf", "FOR_INNEWMARCS", "FOR_HYDRO2", "FOR_PFANT",
           "FOR_NULBAD", "IdMaker", "SID", "translate_sequence"]


import shutil
import os
import logging
import subprocess
from threading import Lock
import a99
import pyfant


# Indexes to use in Conf.sequence property
FOR_INNEWMARCS = 0
FOR_HYDRO2 = 1
FOR_PFANT = 2
FOR_NULBAD = 3

_sequence_dict = {"innewmarcs": FOR_INNEWMARCS, "hydro2": FOR_HYDRO2, "pfant": FOR_PFANT, "nulbad": FOR_NULBAD}

def translate_sequence(sequence):
    """
    Convenience/tolerance function to convert strings with executable names to FOR_*

    Args:
        sequence: examples:
            ["innewmarcs", "hydro2", "pfant", "nulbad"]
            [FOR_INNEWMARCS, FOR_HYDRO2, FOR_PFANT, FOR_NULBAD]
            [True, True, True, True]
            1111
            "1111"
            (all these examples evaluate to [FOR_INNEWMARCS, FOR_HYDRO2, FOR_PFANT, FOR_NULBAD])
    """

    if isinstance(sequence, int):
        sequence = f"{sequence:04}"

    if isinstance(sequence, str):
        ret = [i for i, b in enumerate(sequence) if int(b)]
    else:
        ret = [_sequence_dict[x] if isinstance(x, str) else i if isinstance(x, bool) else x
               for i, x in enumerate(sequence) if not isinstance(x, bool) or x]
    return ret


@a99.froze_it
class SID(object):
    """
    SID -- Session Id and Directory

    Part of code that finds new session id and creates corresponding directory
    """

    @property
    def id(self):
        """
        Information used to create a session directory and to identify this directory.

        When this property is set, the session directory
        (probably named "session-<session_id>") is created
        immediately."""
        return self.__id

    @id.setter
    def id(self, x):
        assert self.__id is None, "Session id already set to \"%s\"" % self.__id
        new_dir = self.__id_maker.session_prefix_singular+x
        os.mkdir(new_dir)
        self.__id = str(x)
        self.__dir = new_dir
        # Will raise if directory exists: directory names must be unique.

    @property
    def dir(self):
        """Name made up with a prefix + (the session id)."""
        return self.__dir

    @property
    def flag_split_dirs(self):
        """Split sessions into several directories with 1000 sessions maximum each.

        This option was created to speed up directory access."""
        return self.__flag_split_dirs
    @flag_split_dirs.setter
    def flag_split_dirs(self, x):
        self.__flag_split_dirs = x

    @property
    def id_maker(self):
        """IdMaker instance"""
        return self.__id_maker
    @id_maker.setter
    def id_maker(self, x):
        self.__id_maker = x

    def __init__(self, id_maker):
        assert isinstance(id_maker, IdMaker)
        self.__id_maker = id_maker
        self.__id = None
        self.__dir = None
        self.__flag_split_dirs = False

    def join_with_session_dir(self, fn):
        """Joins self.session_dir with specified filename to make a path."""
        return os.path.join(self.__dir, fn)

    def make_id(self):
        """Finds an id for a new session and creates corresponding
        directory.

        Directory must be created, otherwise two concurrent threads may grab
        the same id.
        """

        # assert self.session_id is None, "Session id already made"

        if self.__id is not None:
            return
        self.__id, self.__dir = \
            self.__id_maker.make_id(self.__flag_split_dirs)

    def clean(self, flag_remove_dir=True):
        """Deletes session directory with all files inside.

        Args:
          flag_remove_dir=True: if not set, will only delete the contents,
            keeping the directory
        """

        if flag_remove_dir:
            logging.debug("About to remove directory '%s'" % self.__dir)
            shutil.rmtree(self.__dir)
        else:
            # reference: http://stackoverflow.com/questions/185936/delete-folder-contents-in-python
            for the_file in os.listdir(self.__dir):
                file_path = os.path.join(self.__dir, the_file)
                try:
                    if os.path.isfile(file_path):
                        os.unlink(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                except:
                    self.logger.exception("Error removing '%s'" % the_file)


# # Section id managhement
# Part of code that finds new session id and creates corresponding directory


class IdMaker(object):
    """
    Thread-safe id maker + directory maker functionality.
    """

    def __init__(self):
        import pyfant
        self.dirs_per_dir = 1000  # only if flag_split_dirs
        self.session_prefix_singular = pyfant.SESSION_PREFIX_SINGULAR
        self.session_prefix_plural = pyfant.SESSION_PREFIX_PLURAL
        # Lock is necessary to make unique session ids
        self.__lock_session_id = Lock()
        self.__i_id = 0

    def make_id(self, flag_split_dirs):
        """Makes session id and creates corresponding directory.

        This routine is thread-safe.
        """
        with self.__lock_session_id:
            while True:
                new_dir = self.get_session_dirname(flag_split_dirs)
                self.__i_id += 1
                if not os.path.isdir(new_dir):
                    new_id = "%d" % self.__i_id
                    break
            os.makedirs(new_dir)
            return new_id, new_dir

    def get_session_dirname(self, flag_split_dirs):
        """Returns string which is the name of a filesystem directory."""
        if not flag_split_dirs:
            return self.session_prefix_singular+str(self.__i_id)
        d0 = self.session_prefix_plural+str(int(self.__i_id//self.dirs_per_dir)*self.dirs_per_dir)
        d1 = self.session_prefix_singular+str(self.__i_id)
        return os.path.join(d0, d1)

@a99.froze_it
class Conf(object):
    """
    Class holds the configuration of an executable.

    Most options are initialized do None. This means that they won't be added to the
    command line.
    """

    @property
    def popen_text_dest(self):
        """(Read-only) File-like object to act as destination to popen text messages."""
        return self.__popen_text_dest

    @property
    def flag_log_console(self):
        """Display Fortran messages in console (default=True)?"""
        return self.__flag_log_console

    @flag_log_console.setter
    def flag_log_console(self, x):
        self.__flag_log_console = x

    @property
    def flag_log_file(self):
        """Log Fortran messages to "<session dir>/fortran.log" (default=True)?"""
        return self.__flag_log_file

    @flag_log_file.setter
    def flag_log_file(self, x):
        self.__flag_log_file = x

    @property
    def flag_output_to_dir(self):
        """Tweak output filenames to be created inside the session directory
        (default=False)?"""
        return self.__flag_output_to_dir
    @flag_output_to_dir.setter
    def flag_output_to_dir(self, x):
        self.__flag_output_to_dir = x

    @property
    def logger(self):
        if not self.__logger:
            self.__logger = a99.get_python_logger()
        return self.__logger

    @logger.setter
    def logger(self, x):
        self.__logger = x

    @property
    def sid(self):
        """SID object"""
        return self.__sid

    @sid.setter
    def sid(self, x):
        self.__sid = x

    @property
    def opt(self):
        """FileOptions object"""
        return self.__opt

    @opt.setter
    def opt(self, x):
        self.__opt = x


    def __init__(self):
        # # Setup flags
        self.__flag_log_console = True
        self.__flag_log_file = True
        self.__flag_output_to_dir = False

        # # DataFile instances
        # See create_data_files() to see what happens when one or more
        # of these objects is set.
        # running the executable.
        # ## FileMain instance
        self.file_main = None
        # ## FileAbonds instance
        self.file_abonds = None
        # ## FileAbonds instance
        self.file_dissoc = None
        # ## FileHmap instance
        self.file_hmap = None
        # ## FileAtoms instance
        self.file_atoms = None
        # ## FileMolecules instance
        self.file_molecules = None
        # ## FileMollist instance
        self.file_mollist = None

        # # Command-line options
        self.__opt = pyfant.FileOptions()

        # # Read-only properties
        self.__popen_text_dest = None
        self.__logger = None
        self.__sid = SID(_conf_id_maker())


        # # Internals
        self.__flag_configured_before = False

    def configure(self, sequence):
        """Series of configuration actions to take before Runnable can run.

        Runnable can be re-run (useful if fails), so configure() will know
        if it has been called before and skip most operations if so.
        """
        if not self.__flag_configured_before:
            self.__logger = a99.get_python_logger()
            self.__sid.make_id()
            if self.__flag_output_to_dir:
                self.__rename_outputs(sequence)
            if FOR_PFANT in sequence:
                self.__opt.fn_progress = self.__sid.join_with_session_dir("progress.txt")  # this is for pfant

        # Always gotta open logs
        if self.__flag_log_file:
            log_path = self.__sid.join_with_session_dir("fortran.log")
            if self.__flag_log_console:
                stdout_ = a99.LogTwo(log_path)
            else:
                stdout_ = open(log_path, "w")
        else:
            if self.__flag_log_console:
                stdout_ = subprocess.STD_OUTPUT_HANDLE
            else:
                stdout_ = None
        self.__popen_text_dest = stdout_


        self.__create_data_files()
        self.__flag_configured_before = True

    def close_popen_text_dest(self):
        """Closes self.popen_text_dest opened in configure().

        To be called after the Runnable runs."""
        if self.__popen_text_dest is not None:
            self.__popen_text_dest.close()

    def get_file_main(self, opt=None):
        """Tries to figure out a FileMain from current context (i.e., directory contents and command-line options)

        Precedence: self.file_main > (file specified by opt.fn_main) > (f311's default data file)

        Args:
            opt: FileOptions or None

        Returns:
            FileMain or None
        """
        return self.get_file_generic("main", opt)

    def get_file_abonds(self, opt=None):
        """Analogous to get_file_main()"""
        return self.get_file_generic("abonds", opt)

    def get_file_hmap(self, opt=None):
        """Analogous to get_file_main()"""
        return self.get_file_generic("hmap", opt)

    def get_file_generic(self, name, opt=None):
        """
        Returns new File<sth> object, depending on name

        Args:
            name: string
            opt: FileOptions object
        """
        if opt is None:
            opt = self.__opt
        selfattr = self.__getattribute__("file_"+name)
        if selfattr is not None:
            return selfattr

        optattr = opt.__getattribute__("fn_"+name)
        file_ = getattr(pyfant, "File"+name.capitalize())()
        if optattr is None:
            file_.load()  # will try to load default file
        else:
            file_.load(optattr)
        return file_

    def get_flprefix(self, _flag_skip_opt=False, opt=None):
        """Returns the prefix for a pfant output file.

        The returned prefix is used to compose a pfant output file name, e.g.
        <prefix>.norm

        Prefix is looked for in the following locations (in this order):
          1) command-line option, i.e., opt.flprefix; if None,
          2) self.file_main.flprefix; if self.file_main is None,
          3) tries to open the main configuration file
        """
        if opt is None:
            opt = self.__opt
        if not _flag_skip_opt and opt.flprefix is not None:
            return opt.flprefix
        if self.file_main is not None:
            return self.file_main.flprefix
        file_ = self.get_file_main()
        return file_.flprefix

    def get_fwhm(self, opt=None):
        """Returns FWHM from command-line option if present, otherwise from main file"""

        if opt is None:
            opt = self.__opt
        if opt.fwhm is not None:
            return opt.fwhm
        if self.file_main is not None:
            return self.file_main.fwhm
        file_ = self.get_file_main()
        return file_.fwhm

    def get_pfant_output_filepath(self, type_="norm"):
        """Returns path to a pfant output filename.

        Args:
          type_: "spec", "norm", or "cont"

        Looks for this information in several places; see get_flprefix()
        for more information.
        """
        valid_types = ("spec", "norm", "cont")
        assert type_ in valid_types, "type must be in %s" % (valid_types,)
        return self.get_flprefix()+"."+type_

    def get_nulbad_output_filepath(self):
        """Returns path to nulbad output filename.

        Reproduces nulbad logic in determining its output filename, i.e.,
          1) uses --fn_cv if present; if not,
          2) gets flprefix from main configuration file and adds
             (".norm" or ".spec") + ".nulbad." + <fwhm in 5.3 format>
        """
        if self.__opt.fn_cv is not None:
            filename = self.__opt.fn_cv
        else:
            flprefix = self.get_flprefix()
            # True or None evaluates to "norm"
            ext = ("spec" if self.__opt.norm == False else "norm") + \
                  (".nulbad.{:5.3f}".format(self.get_fwhm()))
            filename = flprefix+"."+ext
        return filename

    def get_fn_modeles(self):
        """Returns name of atmospheric model file."""
        return pyfant.FileModBin.default_filename if self.__opt.fn_modeles is None \
         else self.__opt.fn_modeles

    def get_fn_molecules(self):
        """Returns name of molecular lines file"""
        return pyfant.FileMolecules.default_filename if self.__opt.fn_molecules is None \
         else self.__opt.fn_molecules

    def get_args(self):
        """
        Returns a list of command-line arguments (only options that have been set)
        """
        return self.__opt.get_args()

    def rename_outputs(self, sequence, opt=None, sid=None):
        """
        Adds session dir to names of files that will be created by any of the
        executables. To be called *before* create_data_files

        Args:
            sequence: list containing one or more i_* values such as innewmarcs etc.
            opt:
            sid:

        Note: in order to link pfant->nulbad correctly,
              nulbad "--fn_flux" option will not be used.
        """
        self.__rename_outputs(sequence, opt, sid)


    def __create_data_files(self):
        """
        Creates files for all self.file_* that are not None.

        Corresponding self.opt.fn_xxxx will be overwritten.
        """
        for attr_name in dir(self):
            if attr_name.startswith("file_"):
                obj = self.__getattribute__(attr_name)

                if obj is not None:
                    fn_attr_name = "fn_"+attr_name[5:]
                    curr_fn = self.__opt.__getattribute__(fn_attr_name)
                    # Tries to preserve custom file name given to file
                    # (file name only, not directories)
                    base_fn = os.path.basename(curr_fn) if curr_fn \
                     else obj.default_filename
                    new_fn = self.__sid.join_with_session_dir(base_fn)
                    # Saves file
                    obj.save_as(new_fn)
                    # Overwrites config option
                    self.__opt.__setattr__("fn_"+attr_name[5:], new_fn)

    def __rename_outputs(self, sequence, opt=None, sid=None):
        if opt is None:
            opt = self.__opt
        if sid is None:
            sid = self.__sid
        if FOR_INNEWMARCS in sequence:
            # # innewmarcs -> (hydro2, pfant)
            opt.fn_modeles = sid.join_with_session_dir(
             os.path.basename(opt.fn_modeles) if opt.fn_modeles is not None
             else pyfant.FileModBin.default_filename)
            opt.fn_opa = sid.join_with_session_dir(
             os.path.basename(opt.fn_opa) if opt.fn_opa is not None
             else pyfant.FileOpa.default_filename)

        if FOR_HYDRO2 in sequence:
            # # hydro2 -> pfant
            # Task here is to add session_dir to all hydrogen lines filenames, e.g.,
            # if it was "thalpha" it will be something like "session123456/thalpha"

            if not self.file_hmap:
                # if self doesn't have a Hmap object, will load from file
                o = self.file_hmap = pyfant.FileHmap()
                fn = opt.fn_hmap if opt.fn_hmap is not None else \
                 pyfant.FileHmap.default_filename
                o.load(fn)
            else:
                o = self.file_hmap
            for row in o.rows:
                # directory/filename is put between single quotes.
                #
                # Fortran reads this correctly. If not put between quotes, Fortran thinks that the "/"
                # denotes the end of the string.
                row.fn = "'"+sid.join_with_session_dir(row.fn)+"'"
                # Done! new hmap file will be created by create_data_files

        # ** pfant -> nulbad
        if FOR_PFANT in sequence or FOR_NULBAD in sequence:
            flprefix = self.get_flprefix(opt=opt)
            opt.flprefix = sid.join_with_session_dir(flprefix)

        if FOR_NULBAD in sequence:
            opt.fn_flux = None  # will cause nulbad to use flprefix
            if opt.fn_cv:
                # Note that this is recursive, but Combo is meant to be
                # run only once.
                opt.fn_cv = sid.join_with_session_dir(opt.fn_cv)


__conf_id_maker = None

def _conf_id_maker():
    """Returns internal ID maker"""
    global __conf_id_maker
    if __conf_id_maker is None:
        __conf_id_maker = IdMaker()
    return __conf_id_maker




#!/usr/bin/env python
"""
Tunes the "zinf" parameter for each atomic line in atomic lines file

The "zinf" parameter is a distance in angstrom from the centre of an atomic
line. It specifies the calculation range for the line:
[centre-zinf, centre+zinf].

This script runs pfant for each atomic line to determine the width of each
atomic line and thus zinf.

Note: if input files main.dat and abonds.dat are not present in current directory, will copy the
      Sun data.

Note: if modeles.mod is not present in current directory, will run innewmarcs to generate this file.

Note: the precision in the zinf found depends on the calculation step ("pas")
specified in main.dat. A higher "pas" means lower precision and a tendency to
get higher zinf's. This is really not critical. pas=0.02 or pas=0.04 should do.

"""

import argparse
import logging
import copy
import numpy as np
import os.path
import a99
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True
a99.flag_log_console = True

# Fraction of peak to be considered its "settlement"
EPSILON = 1e-4

DEFOUT = "<made-up filename>"


def _get_zinf(lambda_centre, norm):
    """Returns zinf given *normalized* spectrum, or 0 if line if flat.

    Args:
      lambda_centre: centre of line
      norm: Spectrum instance

    Spectrum is expected to contain one line only.
    """

    y = norm.y  # normalized spectrum: values between 0 and 1
    if np.all(np.diff(y) == 0):
        # flat line
        return 0
    ymax = 1
    ymin = min(y)
    i_zinf = np.argmax(norm.y < ymax-(ymax-ymin)*EPSILON)
    zinf = lambda_centre-norm.x[i_zinf]
    return zinf


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('--min', type=float, nargs='?', default=.1,
     help='minimum zinf. If zinf found for a particular line is smaller than '
          'this value, this value will be used instead')
    parser.add_argument('--max', type=float, nargs='?', default=50.,
     help='maximum zinf. If zinf found for a particular line is greater than '
          'this value, this value will be used instead')
    parser.add_argument('--inflate', type=float, nargs='?', default=1.1,
     help='Multiplicative constant to apply a "safety margin". '
          'Each zinf found will be multiplied by this value. For example '
          'a value of INFLATE=1.1 means that all the zinf\'s saved will be 10 '
          'percent larger than those calculated')
    parser.add_argument('--ge_current', action="store_true",
     help='"Greater or Equal to current": If this option is set, the current '
          'zinf in the atomic lines file is used as a lower boundary.')
    parser.add_argument('fn_input', type=str, help='input file name', nargs=1)
    parser.add_argument('fn_output', type=str, help='output file name',
     nargs="?", default=DEFOUT)
    parser.add_argument('--no_clean', action="store_true",
    help='If set, will not remove the session directories.')

    args = parser.parse_args()
    logger = a99.get_python_logger()

    fn_out = args.fn_output
    if fn_out == DEFOUT:
        temp = args.fn_input[0].replace("atoms-untuned-", "")
        fn_out = "atoms-"+temp
        logger.info("Made up output filename: '%s'" % fn_out)

    logger.info("Using minimum zinf = %g" % args.min)
    logger.info("Using MAXIMUM zinf = %g" % args.max)
    logger.info("Using inflate = %g" % args.inflate)
    logger.info("Using ge_current = %s" % args.ge_current)

    file_atoms = pyfant.FileAtoms()
    file_atoms.load(args.fn_input[0])

    logger.info("Number of lines in file '%s': %d" % \
     (args.fn_input[0], file_atoms.num_lines))

    # # Tries to get input files together
    pyfant.setup_inputs(h=False, atoms=False, molecules=False, opa=False)


    # ## Runs innewmarcs if modeles.mod not present
    
    if not os.path.isfile('modeles.mod'):
        logger.info("'modeles.mod' does not exist, running innewmarcs...")
        inn = pyfant.Innewmarcs()
        inn.conf.opt.opa = False
        inn.run()



    # ## As suggested by M. Trevisan, adds 0.5 to all abundances, except EXCEPTIONS
    K_ADD = 0.5
    EXCEPTIONS = ["FE", "HE"]
    logger.info("Preparing abundances (Adding %g to all abundances, " % K_ADD)
    logger.info("except those of %s)..." % str(EXCEPTIONS))
    fa = pyfant.FileAbonds()
    fa.load("abonds.dat")
    cnt = 0
    for i in range(len(fa)):
       if fa.ele[i] not in EXCEPTIONS:
           fa.abol[i] += K_ADD
           cnt += 1
    logger.info("%d abundance(s) changed" % cnt)


    # This is old code, with a minimum value for the abundances but I don't think it is very reallistic
    # MIN_ABOND = 3
    # logger.info("Preparing abundances file with all abundances >= %g ..." % MIN_ABOND)
    # fa = FileAbonds()
    # fa.load("abonds.dat")
    # cnt = 0
    # for i in range(len(fa)):
    #     if fa.abol[i] < MIN_ABOND:
    #         fa.abol[i] = MIN_ABOND
    #         cnt += 1
    # logger.info("%d abundance(s) changed" % cnt)


    # ## Creates the Runnable instances
    PAS = 0.04
    logger.info("Calculation step will be set to %g" % PAS)
    logger.info("Preparing pfant's...")
    combos, ll, i = [], [], 0
    ii = 0
    n = file_atoms.num_lines
    for atom in file_atoms.atoms:
        for line in atom.lines:
            a = copy.copy(atom)  # shallow copy
            a.lines = [line]
            f = pyfant.FileAtoms()
            f.atoms = [a]
            combo = pyfant.Combo([pyfant.FOR_PFANT])
            # Fortran messages will not be displayed in terminal
            combo.conf.flag_log_console = False
            combo.conf.sid.flag_split_dirs = True
            # Forces pfant to run one single iteration (ikeytot will be =1)
            combo.conf.opt.aint = args.max+10
            combo.conf.file_atoms = f
            combo.conf.file_abonds = fa
            combo.conf.opt.logging_level = "warning"
            combo.conf.opt.pas = PAS
            combo.conf.flag_output_to_dir = True
            combo.conf.opt.zinf = args.max
            combo.conf.opt.no_molecules = True
            combo.conf.opt.opa = False
            combo.conf.opt.no_h = True
            # Note that half of the line (needs to)/(will be) calculated
            combo.conf.opt.llzero = line.lambda_-args.max
            combo.conf.opt.llfin = line.lambda_
            combos.append(combo)
            ll.append(a)

            i += 1
            ii += 1
            if ii == 1000:
                print(a99.format_progress(i, n))
                ii = 0

    # # Runs pfant
    logger.info("Running pfant's...")






    # FLAG_CLEAN = True
    # rm = RunnableManager(flag_auto_clean=FLAG_CLEAN)






    rm = pyfant.RunnableManager()
    rm.add_runnables(combos)
    app = a99.get_QApplication([])
    form = pyfant.XRunnableManager(None, rm)
    form.show()
    # it is good to start the manager as late as possible, otherwise
    # the program will hang if, for example, the form fails to be created.
    rm.start()
    app.exec_()
    rm.exit()

    # # Saves log
    LOG_FILENAME = "tune-zinf-status.log"
    with open(LOG_FILENAME, "w") as h:
      h.write(str(rm))
    logger.info("Final status saved to file '%s'" % LOG_FILENAME)

    # # Finds zinf's and creates output file
    flag_success, flag_clean = False, False
    if not rm.flag_finished:
        logger.info("Not finished")
    elif rm.num_failed > 0:
        logger.info("Number of failed tasks: %d" % rm.num_failed)
        logger.info("Please check log files inside directories of sessions that failed")
        logger.info("(session directories will not be removed).")
    else:
        logger.info("Finding zinf's, please wait...")
        # # Calculates zinf and save new atomic lines file
        n = len(combos)
        X = np.zeros((n, 3))  # [algf, kiex, zinf], ...]
        ii = 0
        print(a99.format_progress(0, n))
        cnt_min = 0
        cnt_max = 0
        for i, (a, combo) in enumerate(zip(ll, combos)):
            combo.pfant.load_result()


            # if not FLAG_CLEAN:
            #     combo.pfant.load_result()
            # else:
            #     pass  # result will be already loaded



            norm = combo.pfant.norm  # normalized spectrum
            line = a.lines[0]  # a is single-line FileAtoms object
            zinf = _get_zinf(line.lambda_, norm)
            zinf *= args.inflate
            if zinf == 0:
                # Note that some lines don't appear at all, so there is no way
                # to determine zinf
                logger.warning("zinf is ZERO: (%s) (%s) (%s)" %
                 (combo.conf.sid.dir, a.one_liner_str(), line.one_liner_str()))
            if zinf < args.min:
                zinf = args.min
                cnt_min += 1
            if zinf > args.max:
                # print " %gOLHOLHO (%s) (%s) (%s)" % \
                #  (zinf, combo.conf.session_dir, a.one_liner_str(), line.one_liner_str())
                zinf = args.max
                cnt_max += 1
            if args.ge_current and line.zinf > zinf:
                zinf = line.zinf
            line.zinf = zinf

            i += 1
            ii += 1
            if ii == 100:
                print(a99.format_progress(i, n))
                ii = 0
        print(a99.format_progress(n, n))
        logger.info("zinf's clipped to minimum: %d" % cnt_min)
        logger.info("zinf's clipped to maximum: %d" % cnt_max)

        logger.info("Creating output file...")
        file_atoms.save_as(fn_out)
        flag_clean = not args.no_clean
        flag_success = True


    # # Removes session-* directories
    if flag_clean:
        # this part removes sessionxxxx/sessionyyyy
        logger.info("Cleaning...")

        for i, combo in enumerate(combos):
            try:
                combo.sid.clean()
            except Exception as E:
                logger.info("Error cleaning session for %s: %s" % (combo.name, str(E)))
        # this part removes sessionxxxxx
        # but also removes previous sessions
        # not gonna do it dd = glob.glob("session*")

    if flag_success:
        logger.info("Successfully created file '%s' !!" % fn_out)
